"""
This tests tiled's writing routes with an in-memory store.

Persistent stores are being developed externally to the tiled package.
"""

import base64
from datetime import datetime

import awkward
import dask.dataframe
import numpy
import pandas
import pandas.testing
import pyarrow
import pytest
import sparse
from pandas.testing import assert_frame_equal
from starlette.status import (
    HTTP_404_NOT_FOUND,
    HTTP_409_CONFLICT,
    HTTP_415_UNSUPPORTED_MEDIA_TYPE,
    HTTP_422_UNPROCESSABLE_ENTITY,
)

from ..catalog import in_memory
from ..catalog.adapter import CatalogContainerAdapter
from ..client import Context, from_context, record_history
from ..client.utils import ClientError
from ..mimetypes import PARQUET_MIMETYPE
from ..queries import Key
from ..server.app import build_app
from ..structures.array import BuiltinDtype
from ..structures.core import Spec, StructureFamily
from ..structures.data_source import DataSource
from ..structures.sparse import COOStructure
from ..structures.table import TableStructure
from ..utils import APACHE_ARROW_FILE_MIME_TYPE, patch_mimetypes
from ..validation_registration import ValidationRegistry
from .utils import fail_with_status_code

validation_registry = ValidationRegistry()
validation_registry.register("SomeSpec", lambda *args, **kwargs: None)


@pytest.fixture
def tree(tmpdir):
    return in_memory(
        writable_storage=[
            f"file://localhost{str(tmpdir / 'data')}",
            f"duckdb:///{tmpdir / 'data.duckdb'}",
        ]
    )


def test_write_array_full(tree):
    with Context.from_app(
        build_app(tree, validation_registry=validation_registry)
    ) as context:
        client = from_context(context)

        a = numpy.ones((5, 7))

        metadata = {"scan_id": 1, "method": "A"}
        specs = [Spec("SomeSpec")]
        with record_history() as history:
            client.write_array(a, metadata=metadata, specs=specs)
        # one request for metadata, one for data
        assert len(history.requests) == 1 + 1

        results = client.search(Key("scan_id") == 1)
        result = results.values().first()
        result_array = result.read()

        numpy.testing.assert_equal(result_array, a)
        assert result.metadata == metadata
        assert result.specs == specs


def test_write_large_array_full(tree):
    "Test that a large array is chunked"
    with Context.from_app(
        build_app(tree, validation_registry=validation_registry)
    ) as context:
        client = from_context(context)

        a = numpy.ones(100, dtype=numpy.uint8)
        # Low the limit so we can test on small data, for speed.
        original = client._SUGGESTED_MAX_UPLOAD_SIZE
        client._SUGGESTED_MAX_UPLOAD_SIZE = a.nbytes - 1
        try:
            metadata = {"scan_id": 1, "method": "A"}
            specs = [Spec("SomeSpec")]
            with record_history() as history:
                client.write_array(a, metadata=metadata, specs=specs)
            # one request for metadata, more than one for data
            assert len(history.requests) > 1 + 1

            results = client.search(Key("scan_id") == 1)
            result = results.values().first()
            result_array = result.read()

            numpy.testing.assert_equal(result_array, a)
            assert result.metadata == metadata
            assert result.specs == specs
        finally:
            client._SUGGESTED_MAX_UPLOAD_SIZE = original


def test_write_array_chunked(tree):
    with Context.from_app(
        build_app(tree, validation_registry=validation_registry)
    ) as context:
        client = from_context(context)

        a = dask.array.arange(1500).reshape((50, 30)).rechunk((20, 15))

        metadata = {"scan_id": 1, "method": "A"}
        specs = [Spec("SomeSpec")]
        with record_history() as history:
            client.write_array(a, metadata=metadata, specs=specs)
        # one request for metadata, multiple for data
        assert len(history.requests) == 1 + a.npartitions

        results = client.search(Key("scan_id") == 1)
        result = results.values().first()
        result_array = result.read()

        numpy.testing.assert_equal(result_array, a.compute())
        assert result.metadata == metadata
        assert result.specs == specs


def test_extend_array(tree):
    "Extend an array with additional data, expanding its shape."
    with Context.from_app(
        build_app(tree, validation_registry=validation_registry)
    ) as context:
        client = from_context(context)

        a = numpy.ones((3, 2, 2))
        new_data = numpy.ones((1, 2, 2)) * 2
        full_array = numpy.concatenate((a, new_data), axis=0)

        # Upload a (3, 2, 2) array.
        ac = client.write_array(a)
        assert ac.shape == a.shape

        # Patching data into a region beyond the current extent of the array
        # raises a ValueError (catching a 409 from the server).
        with pytest.raises(ValueError):
            ac.patch(new_data, offset=(3,))
        # With extend=True, the array is expanded.
        ac.patch(new_data, offset=(3,), extend=True)
        # The local cache of the structure is updated.
        assert ac.shape == full_array.shape
        actual = ac.read()
        # The array has the expected shape and data.
        assert actual.shape == full_array.shape
        numpy.testing.assert_equal(actual, full_array)

        # Overwrite data (do not extend).
        revised_data = numpy.ones((1, 2, 2)) * 3
        revised_array = full_array.copy()
        revised_array[3, :, :] = 3
        ac.patch(revised_data, offset=(3,))
        numpy.testing.assert_equal(ac.read(), revised_array)

        # Extend out of order.
        ones = numpy.ones((1, 2, 2))
        ac.patch(ones * 7, offset=(7,), extend=True)
        ac.patch(ones * 5, offset=(5,), extend=True)
        ac.patch(ones * 6, offset=(6,), extend=True)
        numpy.testing.assert_equal(ac[5:6], ones * 5)
        numpy.testing.assert_equal(ac[6:7], ones * 6)
        numpy.testing.assert_equal(ac[7:8], ones * 7)

        # Offset given as an int is acceptable.
        ac.patch(ones * 8, offset=8, extend=True)
        numpy.testing.assert_equal(ac[8:9], ones * 8)

        # Data type must match.
        with pytest.raises(ValueError):
            ac.patch(ones.astype("uint8"), offset=9, extend=True)


def test_write_dataframe_full(tree):
    with Context.from_app(
        build_app(tree, validation_registry=validation_registry)
    ) as context:
        client = from_context(context)

        data = {f"Column{i}": (1 + i) * numpy.ones(5) for i in range(5)}
        df = pandas.DataFrame(data)
        metadata = {"scan_id": 1, "method": "A"}
        specs = [Spec("SomeSpec")]

        with record_history() as history:
            client.write_dataframe(df, metadata=metadata, specs=specs)
        # one request for metadata, one for data
        assert len(history.requests) == 1 + 1

        results = client.search(Key("scan_id") == 1)
        result = results.values().first()
        result_dataframe = result.read()

        pandas.testing.assert_frame_equal(result_dataframe, df)
        assert result.metadata == metadata
        assert result.specs == specs


def test_write_dataframe_partitioned(tree):
    with Context.from_app(
        build_app(tree, validation_registry=validation_registry)
    ) as context:
        client = from_context(context)

        data = {f"Column{i}": (1 + i) * numpy.ones(10) for i in range(5)}
        df = pandas.DataFrame(data)
        ddf = dask.dataframe.from_pandas(df, npartitions=3)
        metadata = {"scan_id": 1, "method": "A"}
        specs = [Spec("SomeSpec")]

        with record_history() as history:
            client.write_dataframe(ddf, metadata=metadata, specs=specs)
        # one request for metadata, multiple for data
        assert len(history.requests) == 1 + 3

        results = client.search(Key("scan_id") == 1)
        result = results.values().first()
        result_dataframe = result.read()

        pandas.testing.assert_frame_equal(result_dataframe, df)
        assert result.metadata == metadata
        assert result.specs == specs


def test_write_dataframe_dict(tree):
    with Context.from_app(
        build_app(tree, validation_registry=validation_registry)
    ) as context:
        client = from_context(context)

        data = {f"Column{i}": (1 + i) * numpy.ones(5) for i in range(5)}
        df = pandas.DataFrame(data)
        metadata = {"scan_id": 1, "method": "A"}
        specs = [Spec("SomeSpec")]

        with record_history() as history:
            client.write_dataframe(data, metadata=metadata, specs=specs)
        # one request for metadata, one for data
        assert len(history.requests) == 1 + 1

        results = client.search(Key("scan_id") == 1)
        result = results.values().first()
        result_dataframe = result.read()

        pandas.testing.assert_frame_equal(result_dataframe, df)
        assert result.metadata == metadata
        assert result.specs == specs


@pytest.mark.parametrize(
    "coo",
    [
        sparse.COO(
            coords=numpy.array([[2, 5]]), data=numpy.array([1.3, 7.5]), shape=(10,)
        ),
        sparse.COO(
            coords=numpy.array([[0, 1], [2, 3]]),
            data=numpy.array([3.8, 4.0]),
            shape=(4, 4),
        ),
    ],
)
def test_write_sparse_full(tree, coo):
    with Context.from_app(
        build_app(tree, validation_registry=validation_registry)
    ) as context:
        client = from_context(context)
        metadata = {"scan_id": 1, "method": "A"}
        specs = [Spec("SomeSpec")]
        with record_history() as history:
            client.write_sparse(
                coords=coo.coords,
                data=coo.data,
                shape=coo.shape,
                metadata=metadata,
                specs=specs,
            )
        # one request for metadata, one for data
        assert len(history.requests) == 1 + 1

        results = client.search(Key("scan_id") == 1)
        result = results.values().first()
        result_array = result.read()

        numpy.testing.assert_equal(result_array.todense(), coo.todense())
        assert result.metadata == metadata
        assert result.specs == specs


def test_write_sparse_chunked(tree):
    with Context.from_app(
        build_app(tree, validation_registry=validation_registry)
    ) as context:
        client = from_context(context)

        metadata = {"scan_id": 1, "method": "A"}
        specs = [Spec("SomeSpec")]
        N = 5
        with record_history() as history:
            x = client.new(
                "sparse",
                [
                    DataSource(
                        structure=COOStructure(
                            shape=(2 * N,),
                            chunks=((N, N),),
                            data_type=BuiltinDtype.from_numpy_dtype(
                                numpy.dtype("float64")
                            ),
                        ),
                        structure_family="sparse",
                    )
                ],
                metadata=metadata,
                specs=specs,
            )
            x.write_block(coords=[[2, 4]], data=[3.1, 2.8], block=(0,))
            x.write_block(coords=[[0, 1]], data=[6.7, 1.2], block=(1,))

        # one request for metadata, multiple for data
        assert len(history.requests) == 1 + 2

        results = client.search(Key("scan_id") == 1)
        result = results.values().first()
        result_array = result.read()
        assert numpy.array_equal(
            result_array.todense(),
            sparse.COO(
                coords=numpy.array([[2, 4, N + 0, N + 1]]),
                data=numpy.array([3.1, 2.8, 6.7, 1.2]),
                shape=(10,),
            ).todense(),
        )

        # numpy.testing.assert_equal(result_array, sparse.COO(coords=[0, 1, ]))
        assert result.metadata == metadata
        assert result.specs == specs


def test_limits(tree):
    "Test various limits on uploaded metadata."

    MAX_ALLOWED_SPECS = 20
    MAX_SPEC_CHARS = 255

    validation_registry = ValidationRegistry()
    for i in range(101):
        validation_registry.register(f"spec{i}", lambda *args, **kwargs: None)
    validation_registry.register("one_too_many", lambda *args, **kwargs: None)
    validation_registry.register("a" * MAX_SPEC_CHARS, lambda *args, **kwargs: None)
    validation_registry.register(
        "a" * (1 + MAX_SPEC_CHARS), lambda *args, **kwargs: None
    )
    with Context.from_app(
        build_app(tree, validation_registry=validation_registry)
    ) as context:
        client = from_context(context)

        # Up to 20 specs are allowed.
        max_allowed_specs = [f"spec{i}" for i in range(MAX_ALLOWED_SPECS)]
        x = client.write_array([1, 2, 3], specs=max_allowed_specs)
        x.update_metadata(specs=max_allowed_specs)  # no-op
        too_many_specs = max_allowed_specs + ["one_too_many"]
        with fail_with_status_code(HTTP_422_UNPROCESSABLE_ENTITY):
            client.write_array([1, 2, 3], specs=too_many_specs)
        with fail_with_status_code(HTTP_422_UNPROCESSABLE_ENTITY):
            x.update_metadata(specs=too_many_specs)

        # Specs cannot repeat.
        has_repeated_spec = ["spec0", "spec1", "spec0"]
        with fail_with_status_code(HTTP_422_UNPROCESSABLE_ENTITY):
            client.write_array([1, 2, 3], specs=has_repeated_spec)
        with fail_with_status_code(HTTP_422_UNPROCESSABLE_ENTITY):
            x.update_metadata(specs=has_repeated_spec)

        # A given spec cannot be too long.
        max_allowed_chars = ["a" * MAX_SPEC_CHARS]
        client.write_array([1, 2, 3], specs=max_allowed_chars)
        too_many_chars = ["a" * (1 + MAX_SPEC_CHARS)]
        with fail_with_status_code(HTTP_422_UNPROCESSABLE_ENTITY):
            client.write_array([1, 2, 3], specs=too_many_chars)
        with fail_with_status_code(HTTP_422_UNPROCESSABLE_ENTITY):
            x.update_metadata(specs=too_many_chars)


def test_metadata_revisions(tree):
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
        ac = client.write_array([1, 2, 3], key="revise_me")
        assert len(ac.metadata_revisions[:]) == 0
        ac.update_metadata(metadata={"a": 1})
        assert ac.metadata["a"] == 1
        assert client["revise_me"].metadata["a"] == 1
        assert len(ac.metadata_revisions[:]) == 1
        ac.update_metadata(metadata={"a": 2})
        assert ac.metadata["a"] == 2
        assert client["revise_me"].metadata["a"] == 2
        assert len(ac.metadata_revisions[:]) == 2
        ac.metadata_revisions.delete_revision(1)
        assert len(ac.metadata_revisions[:]) == 1
        with fail_with_status_code(HTTP_404_NOT_FOUND):
            ac.metadata_revisions.delete_revision(1)


def test_replace_metadata(tree):
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
        ac = client.write_array([1, 2, 3], key="revise_me_with_replace")
        assert len(ac.metadata_revisions[:]) == 0
        ac.replace_metadata(metadata={"a": 1})
        assert ac.metadata["a"] == 1
        assert client["revise_me_with_replace"].metadata["a"] == 1
        assert len(ac.metadata_revisions[:]) == 1
        ac.replace_metadata(metadata={"a": 2})
        assert ac.metadata["a"] == 2
        assert client["revise_me_with_replace"].metadata["a"] == 2
        assert len(ac.metadata_revisions[:]) == 2
        ac.metadata_revisions.delete_revision(1)
        assert len(ac.metadata_revisions[:]) == 1
        with fail_with_status_code(HTTP_404_NOT_FOUND):
            ac.metadata_revisions.delete_revision(1)


def test_drop_revision(tree):
    key = "test_drop_revision"
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
        # Set metadata color=blue.
        ac = client.write_array([1, 2, 3], metadata={"color": "blue"}, key=key)
        assert ac.metadata["color"] == "blue"
        assert client[key].metadata["color"] == "blue"
        assert len(ac.metadata_revisions[:]) == 0
        # Update metadata to color=red, but drop revision.
        ac.update_metadata(metadata={"color": "red"}, drop_revision=True)
        # Metadata is updated; no revision is saved.
        assert ac.metadata["color"] == "red"
        assert client[key].metadata["color"] == "red"
        assert len(ac.metadata_revisions) == 0


def test_merge_patching(tree):
    "Test merge patching of metadata and specs"
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
        ac = client.write_array([1, 2, 3], metadata={"a": 0, "b": 2}, specs=["spec1"])
        ac.patch_metadata(
            metadata_patch={"a": 1, "c": 3}, content_type=patch_mimetypes.MERGE_PATCH
        )
        assert dict(ac.metadata) == {"a": 1, "b": 2, "c": 3}
        assert ac.specs[0].name == "spec1"
        ac.patch_metadata(
            specs_patch=["spec2"], content_type=patch_mimetypes.MERGE_PATCH
        )
        assert [x.name for x in ac.specs] == ["spec2"]


def test_json_patching(tree):
    "Test json patching of metadata and specs"

    validation_registry = ValidationRegistry()

    for i in range(10):
        validation_registry.register(f"spec{i}", lambda *args, **kwargs: None)

    with Context.from_app(
        build_app(tree, validation_registry=validation_registry)
    ) as context:
        client = from_context(context)
        ac = client.write_array([1, 2, 3], metadata={"a": 0, "b": 2}, specs=["spec1"])
        ac.patch_metadata(
            metadata_patch=[
                {"op": "add", "path": "/c", "value": 3},
                {"op": "replace", "path": "/a", "value": 1},
            ],
            content_type=patch_mimetypes.JSON_PATCH,
        )
        assert dict(ac.metadata) == {"a": 1, "b": 2, "c": 3}
        assert ac.specs[0].name == "spec1"
        ac.patch_metadata(
            specs_patch=[{"op": "add", "path": "/1", "value": "spec2"}],
            content_type=patch_mimetypes.JSON_PATCH,
        )
        assert [x.name for x in ac.specs] == ["spec1", "spec2"]


def test_metadata_with_unsafe_objects(tree):
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
        ac = client.write_array(
            [1, 2, 3],
            metadata={"date": datetime.now(), "array": numpy.array([1, 2, 3])},
            key="x",
        )
        # Local copy matches copy fetched from remote.
        assert ac.metadata == client["x"].metadata


@pytest.mark.asyncio
async def test_delete(tree):
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
        client.write_array(
            [1, 2, 3],
            metadata={"date": datetime.now(), "array": numpy.array([1, 2, 3])},
            key="delete_me",
        )
        nodes_before_delete = (await tree.context.execute("SELECT * from nodes")).all()
        assert len(nodes_before_delete) == 1 + 1  # +1 for the root node
        data_sources_before_delete = (
            await tree.context.execute("SELECT * from data_sources")
        ).all()
        assert len(data_sources_before_delete) == 1
        assets_before_delete = (
            await tree.context.execute("SELECT * from assets")
        ).all()
        assert len(assets_before_delete) == 1

        # Writing again with the same name fails.
        with fail_with_status_code(HTTP_409_CONFLICT):
            client.write_array(
                [1, 2, 3],
                metadata={"date": datetime.now(), "array": numpy.array([1, 2, 3])},
                key="delete_me",
            )

        with pytest.raises(ClientError):
            client.delete_contents("delete_me")  # Cannot easily delete internal data
        client.delete_contents("delete_me", external_only=False)

        nodes_after_delete = (await tree.context.execute("SELECT * from nodes")).all()
        assert len(nodes_after_delete) == 0 + 1  # the root node should still exist
        data_sources_after_delete = (
            await tree.context.execute("SELECT * from data_sources")
        ).all()
        assert len(data_sources_after_delete) == 0
        assets_after_delete = (await tree.context.execute("SELECT * from assets")).all()
        assert len(assets_after_delete) == 0

        # Writing again with the same name works now.
        client.write_array(
            [1, 2, 3],
            metadata={"date": datetime.now(), "array": numpy.array([1, 2, 3])},
            key="delete_me",
        )


@pytest.mark.asyncio
async def test_delete_non_empty_node(tree):
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
        a = client.create_container("a")
        b = a.create_container("b")
        c = b.create_container("c")
        d = c.create_container("d")

        # Cannot delete non-empty nodes
        assert "a" in client
        with fail_with_status_code(HTTP_409_CONFLICT):
            client.delete_contents("a")
        assert "b" in a
        with fail_with_status_code(HTTP_409_CONFLICT):
            a.delete_contents("b")
        assert "c" in b
        with fail_with_status_code(HTTP_409_CONFLICT):
            b.delete_contents("c")
        assert "d" in c
        assert not list(d)  # leaf is empty
        # Delete from the bottom up.
        c.delete_contents("d")
        b.delete_contents("c")
        a.delete_contents("b")
        client.delete_contents("a")


@pytest.mark.asyncio
async def test_write_in_container(tree):
    "Create a container and write a structure into it."
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)

        a = client.create_container("a")
        df = pandas.DataFrame({"a": [1, 2, 3]})
        b = a.write_dataframe(df, key="b")
        b.read()
        a.delete_contents("b", external_only=False)
        client.delete_contents("a")

        a = client.create_container("a")
        arr = numpy.array([1, 2, 3])
        b = a.write_array(arr, key="b")
        b.read()
        a.delete_contents("b", external_only=False)
        client.delete_contents("a")

        a = client.create_container("a")
        coo = sparse.COO(
            coords=numpy.array([[2, 5]]), data=numpy.array([1.3, 7.5]), shape=(10,)
        )
        b = a.write_sparse(coords=coo.coords, data=coo.data, shape=coo.shape, key="b")
        b.read()
        a.delete_contents("b", external_only=False)
        client.delete_contents("a")

        a = client.create_container("a")
        array = awkward.Array(
            [
                [{"x": 1.1, "y": [1]}, {"x": 2.2, "y": [1, 2]}],
                [],
                [{"x": 3.3, "y": [1, 2, 3]}],
            ]
        )
        b = a.write_awkward(array, key="b")
        b.read()
        a.delete_contents("b", external_only=False)
        client.delete_contents("a")


@pytest.mark.asyncio
async def test_bytes_in_metadata(tree):
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
        client.create_container("a", metadata={"test": b"raw_bytes"})
        value = client["a"].metadata["test"]
        assert value.startswith("data:application/octet-stream;base64,")
        label, encoded = value.split(",", 1)
        assert base64.b64decode(encoded) == b"raw_bytes"


@pytest.mark.asyncio
async def test_container_export(tree, buffer):
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)

        a = client.create_container("a")
        a.write_array([1, 2, 3], key="b")
        client.export(buffer, format="application/json")


def test_write_with_specified_mimetype(tree):
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context, include_data_sources=True)
        df = pandas.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})
        structure = TableStructure.from_pandas(df)

        for mimetype in [
            PARQUET_MIMETYPE,
            "text/csv",
            APACHE_ARROW_FILE_MIME_TYPE,
        ]:
            x = client.new(
                "table",
                [
                    DataSource(
                        structure_family=StructureFamily.table,
                        structure=structure,
                        mimetype=mimetype,
                    ),
                ],
            )
            x.write_partition(df, 0)
            x.read()
            x.refresh()
            assert x.data_sources()[0].mimetype == mimetype

        # Specifying unsupported mimetype raises expected error.
        with fail_with_status_code(HTTP_415_UNSUPPORTED_MEDIA_TYPE):
            client.new(
                "table",
                [
                    DataSource(
                        structure_family=StructureFamily.table,
                        structure=structure,
                        mimetype="application/x-does-not-exist",
                    ),
                ],
            )


@pytest.mark.parametrize(
    "orig_file, file_to_append, expected_file",
    [
        (
            {"A": [1, 2, 3], "B": [4, 5, 6]},
            {"A": [11, 12, 13], "B": [14, 15, 16]},
            {"A": [1, 2, 3, 11, 12, 13], "B": [4, 5, 6, 14, 15, 16]},
        ),
        (
            {"A": [1.2, 2.5, 3.7], "B": [4.6, 5.8, 6.9]},
            {"A": [11.2, 12.5, 13.7], "B": [14.6, 15.8, 16.9]},
            {
                "A": [1.2, 2.5, 3.7, 11.2, 12.5, 13.7],
                "B": [4.6, 5.8, 6.9, 14.6, 15.8, 16.9],
            },
        ),
        (
            {"C": ["x", "y"], "D": ["a", "b"]},
            {"C": ["xx", "yy", "zz"], "D": ["aa", "bb", "cc"]},
            {"C": ["x", "y", "xx", "yy", "zz"], "D": ["a", "b", "aa", "bb", "cc"]},
        ),
    ],
)
def test_append_partition(
    tree: CatalogContainerAdapter,
    orig_file: dict,
    file_to_append: dict,
    expected_file: dict,
):
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context, include_data_sources=True)
        orig_table = pyarrow.Table.from_pydict(orig_file)
        table_to_append = pyarrow.Table.from_pydict(file_to_append)

        x = client.create_appendable_table(orig_table.schema, key="x")
        x.append_partition(orig_table, 0)
        x.append_partition(table_to_append, 0)
        assert_frame_equal(x.read(), pandas.DataFrame(expected_file), check_dtype=False)


@pytest.mark.parametrize(
    "table_name, expected",
    [
        (None, None),
        ("valid_table_name", None),
        (
            "_invalid_table_name",
            pytest.raises(ValueError, match=r"Malformed SQL identifier.+"),
        ),
        (
            "invalid-table-name",
            pytest.raises(ValueError, match=r"Malformed SQL identifier.+"),
        ),
        (
            "UPPERCASE_TABLE_NAME",
            pytest.raises(ValueError, match=r"Malformed SQL identifier.+"),
        ),
        (
            "",
            pytest.raises(ValueError, match=r"Malformed SQL identifier.+"),
        ),
    ],
)
def test_create_table_with_custom_name(
    tree: CatalogContainerAdapter,
    table_name: str,
    expected: str,
):
    table = pyarrow.Table.from_arrays([[1, 2, 3]], ["column_name"])
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context, include_data_sources=True)
        if isinstance(expected, type(pytest.raises(ValueError))):
            with expected:
                client.create_appendable_table(table.schema, table_name=table_name)
        else:
            x = client.create_appendable_table(table.schema, table_name=table_name)
            x.append_partition(table, 0)
            assert x.read()["column_name"].to_list() == [1, 2, 3]



from pathlib import Path

import numpy
import pandas
import pytest

from ..adapters.csv import CSVArrayAdapter
from ..catalog import in_memory
from ..client import Context, from_context
from ..server.app import build_app
from ..structures.array import ArrayStructure
from ..structures.core import StructureFamily
from ..structures.data_source import Asset, DataSource, Management
from ..structures.table import TableStructure

rng = numpy.random.default_rng(12345)

df1 = pandas.DataFrame(
    {
        "C": ["red", "green", "blue", "white"],
        "D": [10.0, 20.0, 30.0, 40.0],
        "E": [0, 1, 2, 3],
    }
)

arr1 = rng.integers(0, 255, size=(13, 17), dtype="uint8")
arr2 = rng.random(size=(15, 19), dtype="float64")
df_arr1 = pandas.DataFrame(arr1)
df_arr2 = pandas.DataFrame(arr2)


@pytest.fixture(scope="module")
def tree(tmp_path_factory):
    return in_memory(writable_storage=tmp_path_factory.getbasetemp())


@pytest.fixture(scope="module")
def context(tree):
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
        client.create_container(key="x")
        yield context


@pytest.fixture
def csv_table_file(tmpdir):
    fpath = Path(tmpdir, "table.csv")
    df1.to_csv(fpath, index=False)

    yield str(fpath)


@pytest.fixture
def csv_array1_file(tmpdir):
    fpath = Path(tmpdir, "array_1.csv")
    df_arr1.to_csv(fpath, index=False, header=False)

    yield str(fpath)


@pytest.fixture
def csv_array2_file(tmpdir):
    fpath = Path(tmpdir, "array_2.csv")
    df_arr2.to_csv(fpath, index=False, header=False)

    yield str(fpath)


def test_csv_table(context, csv_table_file):
    client = from_context(context)

    csv_assets = [
        Asset(
            data_uri=f"file://localhost/{csv_table_file}",
            is_directory=False,
            parameter="data_uris",
            num=0,
        )
    ]
    csv_data_source = DataSource(
        mimetype="text/csv;header=present",
        assets=csv_assets,
        structure_family=StructureFamily.table,
        structure=TableStructure.from_pandas(df1),
        management=Management.external,
    )

    client["x"].new(
        structure_family=StructureFamily.table,
        data_sources=[csv_data_source],
        key="table",
    )

    read_df = client["x"]["table"].read()
    assert set(read_df.columns) == set(df1.columns)
    assert (read_df == df1).all().all()


def test_csv_arrays(context, csv_array1_file, csv_array2_file):
    client = from_context(context)

    for key, csv_fpath, arr in zip(
        ("array1", "array2"), (csv_array1_file, csv_array2_file), (arr1, arr2)
    ):
        csv_assets = [
            Asset(
                data_uri=f"file://localhost/{csv_fpath}",
                is_directory=False,
                parameter="data_uris",
                num=0,
            )
        ]
        csv_data_source = DataSource(
            mimetype="text/csv;header=absent",
            assets=csv_assets,
            structure_family=StructureFamily.array,
            structure=ArrayStructure.from_array(arr),
            management=Management.external,
        )

        client["x"].new(
            structure_family=StructureFamily.array,
            data_sources=[csv_data_source],
            key=key,
        )

    read_arr1 = client["x"]["array1"].read()
    assert numpy.array_equal(read_arr1, arr1)

    read_arr2 = client["x"]["array2"].read()
    assert numpy.isclose(read_arr2, arr2).all()


def test_csv_arrays_from_uris(csv_array1_file, csv_array2_file):
    array_adapter = CSVArrayAdapter.from_uris(f"file://localhost/{csv_array1_file}")
    read_arr = array_adapter.read()
    assert numpy.isclose(read_arr, arr1).all()

    array_adapter = CSVArrayAdapter.from_uris(f"file://localhost/{csv_array2_file}")
    read_arr = array_adapter.read()
    assert numpy.isclose(read_arr, arr2).all()



import collections
import collections.abc
import functools
import importlib
import itertools
import time
import warnings
from dataclasses import asdict
from typing import TYPE_CHECKING, Any, Iterable, Optional, Union
from urllib.parse import parse_qs, urlparse

import entrypoints
import httpx
import orjson

from ..adapters.utils import IndexersMixin
from ..iterviews import ItemsView, KeysView, ValuesView
from ..queries import KeyLookup
from ..query_registration import default_query_registry
from ..structures.core import StructureFamily
from ..structures.data_source import DataSource
from ..utils import UNCHANGED, OneShotCachedMap, Sentinel, node_repr, safe_json_dump
from .base import STRUCTURE_TYPES, BaseClient
from .utils import (
    MSGPACK_MIME_TYPE,
    ClientError,
    client_for_item,
    export_util,
    handle_error,
    normalize_specs,
    retry_context,
)

if TYPE_CHECKING:
    import pandas
    import pyarrow


class Container(BaseClient, collections.abc.Mapping, IndexersMixin):
    # This maps the structure_family sent by the server to a client-side object that
    # can interpret the structure_family's structure and content. OneShotCachedMap is used to
    # defer imports.

    # This is populated when the first instance is created.
    STRUCTURE_CLIENTS_FROM_ENTRYPOINTS = None

    @classmethod
    def _discover_entrypoints(cls, entrypoint_name) -> OneShotCachedMap[str, Any]:
        return OneShotCachedMap(
            {
                name: entrypoint.load
                for name, entrypoint in entrypoints.get_group_named(
                    entrypoint_name
                ).items()
            }
        )

    @classmethod
    def discover_clients_from_entrypoints(cls):
        """
        Search the software environment for libraries that register structure clients.

        This is called once automatically the first time Node.from_uri
        is called. It is idempotent.
        """
        if cls.STRUCTURE_CLIENTS_FROM_ENTRYPOINTS is not None:
            # short-circuit
            return
        # The modules associated with these entrypoints will be imported
        # lazily, only when the item is first accessed.
        cls.STRUCTURE_CLIENTS_FROM_ENTRYPOINTS = OneShotCachedMap()
        # Check old name (special_client) and new name (structure_client).
        for entrypoint_name in ["tiled.special_client", "tiled.structure_client"]:
            for name, entrypoint in entrypoints.get_group_named(
                entrypoint_name
            ).items():
                cls.STRUCTURE_CLIENTS_FROM_ENTRYPOINTS.set(name, entrypoint.load)
                DEFAULT_STRUCTURE_CLIENT_DISPATCH["numpy"].set(name, entrypoint.load)
                DEFAULT_STRUCTURE_CLIENT_DISPATCH["dask"].set(name, entrypoint.load)

    def __init__(
        self,
        context,
        *,
        item,
        structure_clients,
        queries=None,
        sorting=None,
        structure=None,
        include_data_sources=False,
    ):
        "This is not user-facing. Use Node.from_uri."

        self.structure_clients = structure_clients
        self._queries = list(queries or [])
        self._queries_as_params = _queries_to_params(*self._queries)
        # If the user has not specified a sorting, give the server the opportunity
        # to tell us the default sorting.
        if sorting:
            self._sorting = sorting
        else:
            # In the Python API we encode sorting as (key, direction).
            # This order-based "record" notion does not play well with OpenAPI.
            # In the HTTP API, therefore, we use {"key": key, "direction": direction}.
            self._sorting = [
                (s["key"], int(s["direction"]))
                for s in (item["attributes"].get("sorting") or [])
            ]
        sorting = sorting or item["attributes"].get("sorting")
        self._sorting_params = {
            "sort": ",".join(
                f"{'-' if item[1] < 0 else ''}{item[0]}" for item in self._sorting
            )
        }
        self._reversed_sorting_params = {
            "sort": ",".join(
                f"{'-' if item[1] > 0 else ''}{item[0]}" for item in self._sorting
            )
        }
        super().__init__(
            context=context,
            item=item,
            structure_clients=structure_clients,
            include_data_sources=include_data_sources,
        )

    def __repr__(self):
        # Display up to the first N keys to avoid making a giant service
        # request. Use _keys_slicer because it is unauthenticated.
        N = 10
        return node_repr(self, self._keys_slice(0, N, direction=1))

    @property
    def sorting(self):
        """
        The current sorting of this Node

        Given as a list of tuples where the first entry is the sorting key
        and the second entry indicates ASCENDING (or 1) or DESCENDING (or -1).
        """
        return list(self._sorting)

    def new_variation(
        self,
        *,
        structure_clients=UNCHANGED,
        queries=UNCHANGED,
        sorting=UNCHANGED,
        **kwargs,
    ):
        """
        Create a copy of this Node, optionally varying some parameters.

        This is intended primarily for internal use and use by subclasses.
        """
        if isinstance(structure_clients, str):
            structure_clients = DEFAULT_STRUCTURE_CLIENT_DISPATCH[structure_clients]
        if structure_clients is UNCHANGED:
            structure_clients = self.structure_clients
        if queries is UNCHANGED:
            queries = self._queries
        if sorting is UNCHANGED:
            sorting = self._sorting
        return super().new_variation(
            structure_clients=structure_clients,
            queries=queries,
            sorting=sorting,
            **kwargs,
        )

    def __len__(self):
        # If the contents of this node was provided in-line, there is an
        # implication that the contents are not expected to be dynamic. Used the
        # count provided in the structure.
        if contents := self.item["attributes"]["structure"]["contents"]:
            return len(contents)

        now = time.monotonic()
        if self._cached_len is not None:
            length, deadline = self._cached_len
            if now < deadline:
                # Used the cached value and do not make any request.
                return length
        link = self.item["links"]["search"]
        for attempt in retry_context():
            with attempt:
                content = handle_error(
                    self.context.http_client.get(
                        link,
                        headers={"Accept": MSGPACK_MIME_TYPE},
                        params={
                            **parse_qs(urlparse(link).query),
                            "fields": "count",
                            **self._queries_as_params,
                            **self._sorting_params,
                        },
                    )
                ).json()
        length = content["meta"]["count"]
        self._cached_len = (length, now + LENGTH_CACHE_TTL)
        return length

    def __length_hint__(self):
        # TODO The server should provide an estimated count.
        # https://www.python.org/dev/peps/pep-0424/
        return len(self)

    def __iter__(self, _ignore_inlined_contents=False):
        # If the contents of this node was provided in-line, and we don't need
        # to apply any filtering or sorting, we can slice the in-lined data
        # without fetching anything from the server.
        contents = self.item["attributes"]["structure"]["contents"]
        if (
            (contents is not None)
            and (not self._queries)
            and ((not self.sorting) or (self.sorting == [("_", 1)]))
            and (not _ignore_inlined_contents)
        ):
            return (yield from contents)
        next_page_url = self.item["links"]["search"]
        while next_page_url is not None:
            for attempt in retry_context():
                with attempt:
                    content = handle_error(
                        self.context.http_client.get(
                            next_page_url,
                            headers={"Accept": MSGPACK_MIME_TYPE},
                            params={
                                **parse_qs(urlparse(next_page_url).query),
                                "fields": "",
                                **self._queries_as_params,
                                **self._sorting_params,
                            },
                        )
                    ).json()
            self._cached_len = (
                content["meta"]["count"],
                time.monotonic() + LENGTH_CACHE_TTL,
            )
            for item in content["data"]:
                yield item["id"]
            next_page_url = content["links"]["next"]

    def __getitem__(self, keys, _ignore_inlined_contents=False):
        # These are equivalent:
        #
        # >>> node['a']['b']['c']
        # >>> node[('a', 'b', 'c')]
        # >>> node['a', 'b', 'c']
        #
        # The last two are equivalent at a Python level;
        # both call node.__getitem__(('a', 'b', 'c')).
        #
        # We elide this into a single request to the server rather than
        # a chain of requests. This is not totally straightforward because
        # of this use case:
        #
        # >>> node.search(...)['a', 'b']
        #
        # which must only return a result if 'a' is contained in the search results.
        if not isinstance(keys, tuple):
            keys = (keys,)
        for key in keys:
            if not isinstance(key, str):
                raise TypeError("Containers can only be indexed by strings")
        keys = tuple("/".join(keys).strip("/").split("/"))  # Remove any slashes
        if self._queries:
            # Lookup this key *within the search results* of this Node.
            key, *tail = keys
            tail = tuple(tail)  # list -> tuple
            params = {
                **_queries_to_params(KeyLookup(key)),
                **self._queries_as_params,
                **self._sorting_params,
            }
            if self._include_data_sources:
                params["include_data_sources"] = True
            link = self.item["links"]["search"]
            for attempt in retry_context():
                with attempt:
                    content = handle_error(
                        self.context.http_client.get(
                            link,
                            headers={"Accept": MSGPACK_MIME_TYPE},
                            params={**parse_qs(urlparse(link).query), **params},
                        )
                    ).json()
            self._cached_len = (
                content["meta"]["count"],
                time.monotonic() + LENGTH_CACHE_TTL,
            )
            data = content["data"]
            if not data:
                raise KeyError(key)
            assert (
                len(data) == 1
            ), "The key lookup query must never return more than one result."
            (item,) = data
            result = client_for_item(
                self.context,
                self.structure_clients,
                item,
                include_data_sources=self._include_data_sources,
            )
            if tail:
                result = result[tail]
        else:
            # Straightforwardly look up the keys under this node.
            # There is no search filter in place, so if it is there
            # then we want it.

            # The server may greedily send nested information about children
            # ("inlined contents") to reduce latency. This is how we handle
            # xarray Datasets efficiently, for example.

            # In a loop, walk the key(s). Use inlined contents if we have it.
            # When we reach a key that we don't have inlined contents for, send
            # out a single request with all the rest of the keys, and break
            # the keys-walking loop. We are effectively "jumping" down the tree
            # to the node of interest without downloading information about
            # intermediate parents.
            for i, key in enumerate(keys):
                item = (self.item["attributes"]["structure"]["contents"] or {}).get(key)
                if item is not None and not _ignore_inlined_contents:
                    # The item was inlined, and we should return it as is
                    result = client_for_item(
                        self.context,
                        self.structure_clients,
                        item,
                        include_data_sources=self._include_data_sources,
                    )
                    return (
                        result[keys[i + 1 :]] if keys[i + 1 :] else result  # noqa: E203
                    )
                else:
                    # The item was not inlined, either because nothing was inlined
                    # or because it was added after we fetched the inlined contents.
                    # Make a request for it.
                    try:
                        self_link = self.item["links"]["self"].rstrip("/")
                        params = {}
                        if self._include_data_sources:
                            params["include_data_sources"] = True
                        link = self_link + "".join(f"/{key}" for key in keys[i:])
                        for attempt in retry_context():
                            with attempt:
                                content = handle_error(
                                    self.context.http_client.get(
                                        link,
                                        headers={"Accept": MSGPACK_MIME_TYPE},
                                        params={
                                            **parse_qs(urlparse(link).query),
                                            **params,
                                        },
                                    )
                                ).json()
                    except ClientError as err:
                        if err.response.status_code == httpx.codes.NOT_FOUND:
                            # If this is a scalar lookup, raise KeyError("X") not KeyError(("X",)).
                            err_arg = keys[i:]
                            if len(err_arg) == 1:
                                (err_arg,) = err_arg
                            raise KeyError(err_arg)
                        raise
                    item = content["data"]
                    break

            result = client_for_item(
                self.context,
                self.structure_clients,
                item,
                include_data_sources=self._include_data_sources,
            )
        return result

    def delete_contents(
        self,
        keys: Optional[Union[str, Iterable[str]]] = None,
        recursive: bool = False,
        external_only: bool = True,
    ) -> "Container":
        """Delete the contents of this Container.

        Parameters
        ----------
        keys : str or list of str
            The key(s) to delete. If a list, all keys in the list will be deleted.
            If None (default), delete all contents.
        recursive : bool, optional
            If True, descend into sub-containers and delete their contents too.
            Defaults to False.
        external_only : bool, optional
            If True, only delete externally-managed data. Defaults to True.
        """

        self._cached_len = None
        if keys is None:
            keys = self.keys()
        keys = [keys] if isinstance(keys, str) else keys
        for key in set(keys):
            for attempt in retry_context():
                with attempt:
                    handle_error(
                        self.context.http_client.delete(
                            f"{self.uri}/{key}",
                            params={
                                "recursive": recursive,
                                "external_only": external_only,
                            },
                        )
                    )

        return self

    # The following two methods are used by keys(), values(), items().

    def _keys_slice(
        self, start, stop, direction, page_size=None, *, _ignore_inlined_contents=False
    ):
        # If the contents of this node was provided in-line, and we don't need
        # to apply any filtering or sorting, we can slice the in-lined data
        # without fetching anything from the server.
        contents = self.item["attributes"]["structure"]["contents"]
        if (
            (contents is not None)
            and (not self._queries)
            and ((not self.sorting) or (self.sorting == [("_", 1)]))
            and (not _ignore_inlined_contents)
        ):
            keys = list(contents)
            if direction < 0:
                keys = list(reversed(keys))
            return (yield from keys[start:stop])
        if direction > 0:
            sorting_params = self._sorting_params
        else:
            sorting_params = self._reversed_sorting_params
        assert start >= 0
        assert (stop is None) or (stop >= 0)
        next_page_url = f"{self.item['links']['search']}?page[offset]={start}"
        if page_size is not None:
            next_page_url += f"&page[limit]={page_size}"
        item_counter = itertools.count(start)
        while next_page_url is not None:
            for attempt in retry_context():
                with attempt:
                    content = handle_error(
                        self.context.http_client.get(
                            next_page_url,
                            headers={"Accept": MSGPACK_MIME_TYPE},
                            params={
                                **parse_qs(urlparse(next_page_url).query),
                                "fields": "",
                                **self._queries_as_params,
                                **sorting_params,
                            },
                        )
                    ).json()
            self._cached_len = (
                content["meta"]["count"],
                time.monotonic() + LENGTH_CACHE_TTL,
            )
            for item in content["data"]:
                yield item["id"]
                if stop is not None and next(item_counter) == stop - 1:
                    return
            next_page_url = content["links"]["next"]

    def _items_slice(
        self, start, stop, direction, page_size=None, *, _ignore_inlined_contents=False
    ):
        # If the contents of this node was provided in-line, and we don't need
        # to apply any filtering or sorting, we can slice the in-lined data
        # without fetching anything from the server.
        contents = self.item["attributes"]["structure"]["contents"]
        if (
            (contents is not None)
            and (not self._queries)
            and ((not self.sorting) or (self.sorting == [("_", 1)]))
            and (not _ignore_inlined_contents)
        ):
            items = list(contents.items())
            if direction < 0:
                items = list(reversed(items))
            for key, item in items[start:stop]:
                yield key, client_for_item(
                    self.context,
                    self.structure_clients,
                    item,
                    include_data_sources=self._include_data_sources,
                )
            return
        if direction > 0:
            sorting_params = self._sorting_params
        else:
            sorting_params = self._reversed_sorting_params
        assert start >= 0
        assert (stop is None) or (stop >= 0)
        next_page_url = f"{self.item['links']['search']}?page[offset]={start}"
        if page_size is not None:
            next_page_url += f"&page[limit]={page_size}"
        item_counter = itertools.count(start)
        while next_page_url is not None:
            params = {
                **parse_qs(urlparse(next_page_url).query),
                **self._queries_as_params,
                **sorting_params,
            }
            if self._include_data_sources:
                params["include_data_sources"] = True
            for attempt in retry_context():
                with attempt:
                    content = handle_error(
                        self.context.http_client.get(
                            next_page_url,
                            headers={"Accept": MSGPACK_MIME_TYPE},
                            params=params,
                        )
                    ).json()
            self._cached_len = (
                content["meta"]["count"],
                time.monotonic() + LENGTH_CACHE_TTL,
            )
            for item in content["data"]:
                key = item["id"]
                yield key, client_for_item(
                    self.context,
                    self.structure_clients,
                    item,
                    include_data_sources=self._include_data_sources,
                )
                if stop is not None and next(item_counter) == stop - 1:
                    return
            next_page_url = content["links"]["next"]

    def keys(self):
        return KeysView(lambda: len(self), self._keys_slice)

    def values(self):
        return ValuesView(lambda: len(self), self._items_slice)

    def items(self):
        return ItemsView(lambda: len(self), self._items_slice)

    def search(self, query):
        """
        Make a Node with a subset of this Node's entries, filtered by query.

        Examples
        --------

        >>> from tiled.queries import FullText
        >>> tree.search(FullText("hello"))
        """
        return self.new_variation(queries=self._queries + [query])

    def distinct(
        self, *metadata_keys, structure_families=False, specs=False, counts=False
    ):
        """
        Get the unique values and optionally counts of metadata_keys,
        structure_families, and specs in this Node's entries

        Examples
        --------

        Query all the distinct values of a key.

        >>> tree.distinct("foo", counts=True)

        Query for multiple keys at once.

        >>> tree.distinct("foo", "bar", counts=True)
        """

        link = self.item["links"]["self"].replace("/metadata", "/distinct", 1)
        for attempt in retry_context():
            with attempt:
                distinct = handle_error(
                    self.context.http_client.get(
                        link,
                        headers={"Accept": MSGPACK_MIME_TYPE},
                        params={
                            **parse_qs(urlparse(link).query),
                            "metadata": metadata_keys,
                            "structure_families": structure_families,
                            "specs": specs,
                            "counts": counts,
                            **self._queries_as_params,
                        },
                    )
                ).json()
        return distinct

    def sort(self, *sorting):
        """
        Make a Node with the same entries but sorted according to `sorting`.

        Examples
        --------

        Sort by "color" in ascending order, and then by "height" in descending order.

        >>> from tiled.client import ASCENDING, DESCENDING
        >>> tree.sort(("color", ASCENDING), ("height", DESCENDING))

        Note that ``1`` may be used as a synonym for ``ASCENDING``, and ``-1``
        may be used as a synonym for ``DESCENDING``.
        """
        return self.new_variation(sorting=sorting)

    def export(self, filepath, fields=None, *, format=None):
        """
        Download metadata and data below this node in some format and write to a file.

        Parameters
        ----------
        file: str or buffer
            Filepath or writeable buffer.
        fields: List[str], optional
            Filter which items in this node to export.
        format : str, optional
            If format is None and `file` is a filepath, the format is inferred
            from the name, like 'table.h5' implies format="application/x-hdf5". The format
            may be given as a file extension ("h5") or a media type ("application/x-hdf5").

        Examples
        --------

        Export all.

        >>> a.export("everything.h5")

        """
        params = {}
        if fields is not None:
            params["field"] = fields
        return export_util(
            filepath,
            format,
            self.context.http_client.get,
            self.item["links"]["full"],
            params=params,
        )

    def _ipython_key_completions_(self):
        """
        Provide method for the key-autocompletions in IPython.

        See http://ipython.readthedocs.io/en/stable/config/integrating.html#tab-completion
        """
        MAX_ENTRIES_SUPPORTED = 40
        try:
            if len(self) > MAX_ENTRIES_SUPPORTED:
                MSG = (
                    "Tab-completion is not supported on this particular Node "
                    "because it has a large number of entries."
                )
                warnings.warn(MSG)
                return []
            else:
                return list(self)
        except Exception:
            # Do not print messy traceback from thread. Just fail silently.
            return []

    def new(
        self,
        structure_family,
        data_sources,
        *,
        key=None,
        metadata=None,
        specs=None,
        access_tags=None,
    ):
        """
        Create a new item within this Node.

        This is a low-level method. See high-level convenience methods listed below.

        See Also
        --------
        write_array
        write_dataframe
        write_coo_array
        """
        self._cached_len = None
        metadata = metadata or {}
        access_blob = {"tags": access_tags} if access_tags is not None else {}

        item = {
            "attributes": {
                "metadata": metadata,
                "structure_family": StructureFamily(structure_family),
                "specs": normalize_specs(specs or []),
                "data_sources": [asdict(data_source) for data_source in data_sources],
                "access_blob": access_blob,
            }
        }
        body = dict(item["attributes"])
        if key is not None:
            body["id"] = key

        # if check:
        if any(data_source.assets for data_source in data_sources):
            endpoint = self.uri.replace("/metadata/", "/register/", 1)
        else:
            endpoint = self.uri

        for attempt in retry_context():
            with attempt:
                document = handle_error(
                    self.context.http_client.post(
                        endpoint,
                        headers={"Accept": MSGPACK_MIME_TYPE},
                        content=safe_json_dump(body),
                    )
                ).json()

        if structure_family == StructureFamily.container:
            structure = {"contents": None, "count": None}
        else:
            # Only containers can have multiple data_sources right now.
            (data_source,) = data_sources
            structure = data_source.structure
        item["attributes"]["structure"] = structure

        # If server returned modified metadata update the local copy.
        # Otherwise, round-trip it through a JSON serializer locally
        # to ensure that any special types (e.g. datetimes, numpy arrays)
        # are normalized to natively JSON serializable types.
        if "metadata" in document:
            item["attributes"]["metadata"] = document.pop("metadata")
        else:
            item["attributes"]["metadata"] = orjson.loads(
                safe_json_dump(item["attributes"]["metadata"])
            )

        # Ditto for structure
        if "structure" in document:
            item["attributes"]["structure"] = STRUCTURE_TYPES[structure_family](
                document.pop("structure")
            )

        # And for data sources
        if "data_sources" in document:
            item["attributes"]["data_sources"] = [
                ds for ds in document.pop("data_sources")
            ]

        # And for access_blob
        if "access_blob" in document:
            item["attributes"]["access_blob"] = document.pop("access_blob")

        # Merge in "id" and "links" returned by the server.
        item.update(document)

        # Ensure this is a dataclass, not a dict.
        # When we apply type hints and mypy to the client it should be possible
        # to dispense with this.
        if (structure_family != StructureFamily.container) and isinstance(
            structure, dict
        ):
            structure_type = STRUCTURE_TYPES[structure_family]
            structure = structure_type.from_json(structure)

        return client_for_item(
            self.context,
            self.structure_clients,
            item,
            structure=structure,
            include_data_sources=self._include_data_sources,
        )

    # When (re)chunking arrays for upload, we use this limit
    # to attempt to avoid bumping into size limits.
    _SUGGESTED_MAX_UPLOAD_SIZE = 100_000_000  # 100 MB

    def create_container(
        self,
        key=None,
        *,
        metadata=None,
        specs=None,
        access_tags=None,
    ):
        """Create a new, empty container.

        Parameters
        ----------
        key : str, optional
            Key (name) for this new node. If None, the server will provide a unique key.
        metadata : dict, optional
            User metadata. May be nested. Must contain only basic types
            (e.g. numbers, strings, lists, dicts) that are JSON-serializable.
        specs : List[Spec], optional
            List of names that are used to label that the data and/or metadata
            conform to some named standard specification.
        access_tags: List[str], optional
            Server-specific authZ tags in list form, used to confer access to the node.

        """
        return self.new(
            StructureFamily.container,
            [],
            key=key,
            metadata=metadata,
            specs=specs,
            access_tags=access_tags,
        )

    def write_array(
        self,
        array,
        *,
        key=None,
        metadata=None,
        dims=None,
        specs=None,
        access_tags=None,
    ):
        """
        EXPERIMENTAL: Write an array.

        Parameters
        ----------
        array : array-like
        key : str, optional
            Key (name) for this new node. If None, the server will provide a unique key.
        metadata : dict, optional
            User metadata. May be nested. Must contain only basic types
            (e.g. numbers, strings, lists, dicts) that are JSON-serializable.
        dims : List[str], optional
            A label for each dimension of the array.
        specs : List[Spec], optional
            List of names that are used to label that the data and/or metadata
            conform to some named standard specification.
        access_tags: List[str], optional
            Server-specific authZ tags in list form, used to confer access to the node.

        """
        import dask.array
        import numpy
        from dask.array.core import normalize_chunks

        from ..structures.array import ArrayStructure, BuiltinDtype

        if not (hasattr(array, "shape") and hasattr(array, "dtype")):
            # This does not implement enough of the array-like interface.
            # Coerce to numpy.
            array = numpy.asarray(array)

        # Determine chunks such that each chunk is not too large to upload.
        # Any existing chunking will be taken into account.
        # If the array is small, there will be only one chunk.
        if hasattr(array, "chunks"):
            chunks = normalize_chunks(
                array.chunks,
                limit=self._SUGGESTED_MAX_UPLOAD_SIZE,
                dtype=array.dtype,
                shape=array.shape,
            )
        else:
            chunks = normalize_chunks(
                tuple("auto" for _ in array.shape),
                limit=self._SUGGESTED_MAX_UPLOAD_SIZE,
                dtype=array.dtype,
                shape=array.shape,
            )

        structure = ArrayStructure(
            shape=array.shape,
            chunks=chunks,
            dims=dims,
            data_type=BuiltinDtype.from_numpy_dtype(array.dtype),
        )
        client = self.new(
            StructureFamily.array,
            [
                DataSource(structure=structure, structure_family=StructureFamily.array),
            ],
            key=key,
            metadata=metadata,
            specs=specs,
            access_tags=access_tags,
        )
        chunked = any(len(dim) > 1 for dim in chunks)
        if not chunked:
            client.write(array)
        else:
            # Fan out client.write_block over each chunk using dask.
            if isinstance(array, dask.array.Array):
                da = array.rechunk(chunks)
            else:
                da = dask.array.from_array(array, chunks=chunks)

            # Dask inspects the signature and passes block_id in if present.
            # It also apparently calls it with an empty array and block_id
            # once, so we catch that call and become a no-op.
            def write_block(x, block_id, client):
                if len(block_id):
                    client.write_block(x, block=block_id)
                return x

            # TODO Is there a fire-and-forget analogue such that we don't need
            # to bother with the return type?
            da.map_blocks(write_block, dtype=da.dtype, client=client).compute()
        return client

    def write_awkward(
        self,
        array,
        *,
        key=None,
        metadata=None,
        dims=None,
        specs=None,
        access_tags=None,
    ):
        """
        Write an AwkwardArray.

        Parameters
        ----------
        array: awkward.Array
        key : str, optional
            Key (name) for this new node. If None, the server will provide a unique key.
        metadata : dict, optional
            User metadata. May be nested. Must contain only basic types
            (e.g. numbers, strings, lists, dicts) that are JSON-serializable.
        dims : List[str], optional
            A label for each dimension of the array.
        specs : List[Spec], optional
            List of names that are used to label that the data and/or metadata
            conform to some named standard specification.
        access_tags: List[str], optional
            Server-specific authZ tags in list form, used to confer access to the node.
        """
        import awkward

        from ..structures.awkward import AwkwardStructure

        packed = awkward.to_packed(array)
        form, length, container = awkward.to_buffers(packed)
        structure = AwkwardStructure(
            length=length,
            form=form.to_dict(),
        )
        client = self.new(
            StructureFamily.awkward,
            [
                DataSource(
                    structure=structure, structure_family=StructureFamily.awkward
                ),
            ],
            key=key,
            metadata=metadata,
            specs=specs,
            access_tags=access_tags,
        )
        client.write(container)
        return client

    def write_sparse(
        self,
        coords,
        data,
        shape,
        *,
        key=None,
        metadata=None,
        dims=None,
        specs=None,
        access_tags=None,
    ):
        """
        EXPERIMENTAL: Write a sparse array.

        Parameters
        ----------
        coords : array-like
        data : array-like
        shape : tuple
        key : str, optional
            Key (name) for this new node. If None, the server will provide a unique key.
        metadata : dict, optional
            User metadata. May be nested. Must contain only basic types
            (e.g. numbers, strings, lists, dicts) that are JSON-serializable.
        dims : List[str], optional
            A label for each dimension of the array.
        specs : List[Spec], optional
            List of names that are used to label that the data and/or metadata
            conform to some named standard specification.
        access_tags: List[str], optional
            Server-specific authZ tags in list form, used to confer access to the node.

        Examples
        --------

        Write a sparse.COO array.

        >>> import sparse
        >>> from numpy import array
        >>> coo = sparse.COO(coords=array([[2, 5]]), data=array([1.3, 7.5]), shape=(10,))
        >>> c.write_sparse(coords=coo.coords, data=coo.data, shape=coo.shape)

        This only supports a single chunk. For chunked upload, use lower-level methods.

        # Define the overall shape and the dimensions of each chunk.
        >>> from tiled.structures.sparse import COOStructure
        >>> structure = COOStructure(shape=(10,), chunks=((5, 5),))
        >>> data_source = DataSource(structure=structure, structure_family=StructureFamily.sparse)
        >>> x = c.new("sparse", [data_source])
        # Upload the data in each chunk.
        # Coords are given with in the reference frame of each chunk.
        >>> x.write_block(coords=array([[2, 4]]), data=array([3.1, 2.8]), block=(0,))
        >>> x.write_block(coords=array([[0, 1]]), data=array([6.7, 1.2]), block=(1,))
        """
        from ..structures.array import BuiltinDtype
        from ..structures.sparse import COOStructure

        structure = COOStructure(
            shape=shape,
            # This method only supports single-chunk COO arrays.
            chunks=tuple((dim,) for dim in shape),
            dims=dims,
            data_type=BuiltinDtype.from_numpy_dtype(data.dtype),
            coord_data_type=BuiltinDtype.from_numpy_dtype(coords.dtype),
        )
        client = self.new(
            StructureFamily.sparse,
            [
                DataSource(
                    structure=structure, structure_family=StructureFamily.sparse
                ),
            ],
            key=key,
            metadata=metadata,
            specs=specs,
            access_tags=access_tags,
        )
        client.write(coords, data)
        return client

    def create_appendable_table(
        self,
        schema: "pyarrow.Schema",
        npartitions: int = 1,
        *,
        key=None,
        metadata=None,
        specs=None,
        access_tags=None,
        table_name: Optional[str] = None,
    ):
        """Initialize a table whose rows can be appended to a partition.

        Parameters
        ----------
        schema : column names and dtypes info in the form of pyarrow.Schema
        npartitions : int, optional
            Number of partitions to create. Default is 1.
        key : str, optional
            Key (name) for this new node. If None, the server will provide a unique key.
        metadata : dict, optional
            User metadata. May be nested. Must contain only basic types
            (e.g. numbers, strings, lists, dicts) that are JSON-serializable.
        specs : List[Spec], optional
            List of names that are used to label that the data and/or metadata
            conform to some named standard specification.
        access_tags: List[str], optional
            Server-specific authZ tags in list form, used to confer access to the node.
        table_name : str, optional
            Optionally provide a name for the table this should be stored in.
            By default a name unique to the schema will be chosen.

        See Also
        --------
        write_dataframe
        """
        parameters = {}
        if table_name is not None:
            parameters["table_name"] = table_name

        from ..structures.table import TableStructure

        structure = TableStructure.from_schema(schema, npartitions)

        client = self.new(
            StructureFamily.table,
            [
                DataSource(
                    structure=structure,
                    structure_family=StructureFamily.table,
                    mimetype="application/x-tiled-sql-table",
                    parameters=parameters,
                )
            ],
            key=key,
            metadata=metadata or {},
            specs=specs or [],
            access_tags=access_tags,
        )

        return client

    def write_dataframe(
        self,
        dataframe: Union["pandas.DataFrame", dict[str, Any]],
        *,
        key=None,
        metadata=None,
        specs=None,
        access_tags=None,
    ):
        """Write a DataFrame.

        Parameters
        ----------
        dataframe : pandas.DataFrame or dict
        key : str, optional
            Key (name) for this new node. If None, the server will provide a unique key.
        metadata : dict, optional
            User metadata. May be nested. Must contain only basic types
            (e.g. numbers, strings, lists, dicts) that are JSON-serializable.
        specs : List[Spec], optional
            List of names that are used to label that the data and/or metadata
            conform to some named standard specification.
        access_tags: List[str], optional
            Server-specific authZ tags in list form, used to confer access to the node.

        See Also
        --------
        create_appendable_table
        """
        import dask.dataframe

        from ..structures.table import TableStructure

        if isinstance(dataframe, dask.dataframe.DataFrame):
            structure = TableStructure.from_dask_dataframe(dataframe)
        elif isinstance(dataframe, dict):
            # table as dict, e.g. {"a": [1,2,3], "b": [4,5,6]}
            structure = TableStructure.from_dict(dataframe)
        else:
            structure = TableStructure.from_pandas(dataframe)

        client = self.new(
            StructureFamily.table,
            [
                DataSource(
                    structure=structure,
                    structure_family=StructureFamily.table,
                )
            ],
            key=key,
            metadata=metadata or {},
            specs=specs or [],
            access_tags=access_tags,
        )

        if hasattr(dataframe, "partitions"):
            if isinstance(dataframe, dask.dataframe.DataFrame):
                ddf = dataframe
            else:
                raise NotImplementedError(
                    f"Unsure how to handle type {type(dataframe)}"
                )

            ddf.map_partitions(
                functools.partial(_write_partition, client=client), meta=dataframe._meta
            ).compute()
        else:
            client.write(dataframe)

        return client


def _queries_to_params(*queries):
    "Compute GET params from the queries."
    params = collections.defaultdict(list)
    for query in queries:
        name = default_query_registry.query_type_to_name[type(query)]
        for field, value in query.encode().items():
            if value is not None:
                params[f"filter[{name}][condition][{field}]"].append(value)
    return dict(params)


LENGTH_CACHE_TTL = 1  # second


class Ascending(Sentinel):
    "Intended for more readable sorting operations. An alias for 1."

    def __index__(self):
        return 1


class Descending(Sentinel):
    "Intended for more readable sorting operations. An alias for -1."

    def __index__(self):
        return -1


ASCENDING = Ascending("ASCENDING")
"Ascending sort order. An alias for 1."
DESCENDING = Descending("DESCENDING")
"Descending sort order. An alias for -1."


class _LazyLoad:
    # This exists because lambdas and closures cannot be pickled.
    def __init__(self, import_module_args, attr_name):
        self.import_module_args = import_module_args
        self.attr_name = attr_name

    def __call__(self):
        return getattr(
            importlib.import_module(*self.import_module_args), self.attr_name
        )


class _Wrap:
    # This exists because lambdas and closures cannot be pickled.
    def __init__(self, obj):
        self.obj = obj

    def __call__(self):
        return self.obj


def _write_partition(x, partition_info, client):
    client.write_partition(x, partition_info["number"])
    return x


DEFAULT_STRUCTURE_CLIENT_DISPATCH = {
    "numpy": OneShotCachedMap(
        {
            "container": _Wrap(Container),
            "composite": _LazyLoad(
                ("..composite", Container.__module__), "CompositeClient"
            ),
            "array": _LazyLoad(("..array", Container.__module__), "ArrayClient"),
            "awkward": _LazyLoad(("..awkward", Container.__module__), "AwkwardClient"),
            "dataframe": _LazyLoad(
                ("..dataframe", Container.__module__), "DataFrameClient"
            ),
            "sparse": _LazyLoad(("..sparse", Container.__module__), "SparseClient"),
            "table": _LazyLoad(
                ("..dataframe", Container.__module__), "DataFrameClient"
            ),
            "xarray_dataset": _LazyLoad(
                ("..xarray", Container.__module__), "DatasetClient"
            ),
        }
    ),
    "dask": OneShotCachedMap(
        {
            "container": _Wrap(Container),
            "composite": _LazyLoad(
                ("..composite", Container.__module__), "CompositeClient"
            ),
            "array": _LazyLoad(("..array", Container.__module__), "DaskArrayClient"),
            # TODO Create DaskAwkwardClient
            # "awkward": _LazyLoad(("..awkward", Container.__module__), "DaskAwkwardClient"),
            "dataframe": _LazyLoad(
                ("..dataframe", Container.__module__), "DaskDataFrameClient"
            ),
            "sparse": _LazyLoad(("..sparse", Container.__module__), "SparseClient"),
            "table": _LazyLoad(
                ("..dataframe", Container.__module__), "DaskDataFrameClient"
            ),
            "xarray_dataset": _LazyLoad(
                ("..xarray", Container.__module__), "DaskDatasetClient"
            ),
        }
    ),
}



from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Set, Tuple, Union

import dask.dataframe
import numpy
import pandas
import pytest
import sparse
from numpy.typing import NDArray
from pytest_mock import MockFixture

from ..access_control.access_policies import ALL_ACCESS
from ..access_control.scopes import ALL_SCOPES
from ..adapters.awkward_directory_container import DirectoryContainer
from ..adapters.protocols import (
    AccessPolicy,
    ArrayAdapter,
    AwkwardAdapter,
    BaseAdapter,
    SparseAdapter,
    TableAdapter,
)
from ..ndslice import NDSlice
from ..server.schemas import Principal, PrincipalType
from ..storage import Storage
from ..structures.array import ArrayStructure, BuiltinDtype
from ..structures.awkward import AwkwardStructure
from ..structures.core import Spec, StructureFamily
from ..structures.sparse import COOStructure
from ..structures.table import TableStructure
from ..type_aliases import JSON, Filters, Scopes


class CustomArrayAdapter:
    structure_family: Literal[StructureFamily.array] = StructureFamily.array
    supported_storage: Set[type[Storage]] = set()

    def __init__(
        self,
        array: NDArray[Any],
        structure: ArrayStructure,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
    ) -> None:
        self._array = array
        self._structure = structure
        self._metadata = metadata or {}
        self._specs = specs or []

    def structure(self) -> ArrayStructure:
        return self._structure

    def read(self, slice: NDSlice) -> NDArray[Any]:
        return self._array

    def read_block(self, block: Tuple[int, ...]) -> NDArray[Any]:
        return self._array

    def specs(self) -> List[Spec]:
        return self._specs

    def metadata(self) -> JSON:
        return self._metadata


def arrayadapter_protocol_functions(
    adapter: ArrayAdapter, slice: NDSlice, block: Tuple[int, ...]
) -> None:
    adapter.structure()
    adapter.read(slice)
    adapter.read_block(block)
    adapter.specs()
    adapter.metadata()


def test_arrayadapter_protocol(mocker: MockFixture) -> None:
    mock_call = mocker.patch.object(CustomArrayAdapter, "structure")
    mock_call2 = mocker.patch.object(CustomArrayAdapter, "read")
    mock_call3 = mocker.patch.object(CustomArrayAdapter, "read_block")
    mock_call4 = mocker.patch.object(CustomArrayAdapter, "specs")
    mock_call5 = mocker.patch.object(CustomArrayAdapter, "metadata")

    structure = ArrayStructure(
        data_type=BuiltinDtype.from_numpy_dtype(numpy.dtype("int32")),
        shape=(2, 512, 512),
        chunks=((1, 1), (512,), (512,)),
        dims=("time", "x", "y"),  # optional
    )

    array = numpy.random.rand(2, 512, 512)
    metadata: JSON = {"foo": "bar"}
    anyslice = NDSlice(1, 1, 1)
    anyblock = NDSlice(1, 1, 1)

    anyarrayadapter = CustomArrayAdapter(array, structure, metadata=metadata)
    assert anyarrayadapter.structure_family == StructureFamily.array

    arrayadapter_protocol_functions(anyarrayadapter, anyslice, anyblock)
    mock_call.assert_called_once()
    mock_call2.assert_called_once_with(anyslice)
    mock_call3.assert_called_once_with(anyblock)
    mock_call4.assert_called_once()
    mock_call5.assert_called_once()


class CustomAwkwardAdapter:
    structure_family: Literal[StructureFamily.awkward] = StructureFamily.awkward
    supported_storage: Set[type[Storage]] = set()

    def __init__(
        self,
        container: DirectoryContainer,
        structure: AwkwardStructure,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
    ) -> None:
        self.container = container
        self._metadata = metadata or {}
        self._structure = structure
        self._specs = list(specs or [])

    def structure(self) -> AwkwardStructure:
        return self._structure

    def read(self) -> NDArray[Any]:
        return numpy.random.rand(4)

    def read_buffers(self, form_keys: Optional[List[str]] = None) -> Dict[str, Any]:
        return {"a": 123}

    def write(self, container: DirectoryContainer) -> None:
        return None

    def specs(self) -> List[Spec]:
        return self._specs

    def metadata(self) -> JSON:
        return self._metadata


def awkwardadapter_protocol_functions(
    adapter: AwkwardAdapter,
    slice: NDSlice,
    form_keys: Optional[List[str]],
    container: DirectoryContainer,
) -> None:
    adapter.structure()
    adapter.read()
    adapter.read_buffers(form_keys)
    adapter.write(container)
    adapter.specs()
    adapter.metadata()


def test_awkwardadapter_protocol(mocker: MockFixture) -> None:
    mock_call = mocker.patch.object(CustomAwkwardAdapter, "structure")
    mock_call2 = mocker.patch.object(CustomAwkwardAdapter, "read")
    mock_call3 = mocker.patch.object(CustomAwkwardAdapter, "read_buffers")
    mock_call4 = mocker.patch.object(CustomAwkwardAdapter, "write")
    mock_call5 = mocker.patch.object(CustomAwkwardAdapter, "specs")
    mock_call6 = mocker.patch.object(CustomAwkwardAdapter, "metadata")

    structure = AwkwardStructure(length=2, form={"a": "b"})

    metadata: JSON = {"foo": "bar"}
    anyslice = NDSlice(1, 1, 1)
    container = DirectoryContainer(directory=Path("somedirectory"), form={})
    form_keys = ["a", "b", "c"]

    anyawkwardadapter = CustomAwkwardAdapter(container, structure, metadata=metadata)

    assert anyawkwardadapter.structure_family == StructureFamily.awkward

    awkwardadapter_protocol_functions(anyawkwardadapter, anyslice, form_keys, container)
    mock_call.assert_called_once()
    mock_call2.assert_called_once()
    mock_call3.assert_called_once_with(form_keys)
    mock_call4.assert_called_once_with(container)
    mock_call5.assert_called_once()
    mock_call6.assert_called_once()


class CustomSparseAdapter:
    structure_family: Literal[StructureFamily.sparse] = StructureFamily.sparse
    supported_storage: Set[type[Storage]] = set()

    def __init__(
        self,
        blocks: Dict[Tuple[int, ...], Tuple[NDArray[Any], Any]],
        structure: COOStructure,
        *,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
    ) -> None:
        """
        Construct from blocks with coords given in block-local reference frame.
        Parameters
        ----------
        blocks :
        structure :
        metadata :
        specs :
        """
        self.blocks = blocks
        self._metadata = metadata or {}
        self._structure = structure
        self._specs = specs or []

        all_coords = [[1], [1]]
        all_data = [[1], [1]]

        self.arr = sparse.COO(
            data=numpy.concatenate(all_data),
            coords=numpy.concatenate(all_coords, axis=-1),
            shape=self._structure.shape,
        )

    def structure(self) -> COOStructure:
        return self._structure

    def read(self, slice: NDSlice) -> sparse.COO:
        return self.arr

    def read_block(self, block: Tuple[int, ...]) -> sparse.COO:
        return self.arr

    def specs(self) -> List[Spec]:
        return self._specs

    def metadata(self) -> JSON:
        return self._metadata


def sparseadapter_protocol_functions(
    adapter: SparseAdapter, slice: NDSlice, block: Tuple[int, ...]
) -> None:
    adapter.structure()
    adapter.read(slice)
    adapter.read_block(block)
    adapter.specs()
    adapter.metadata()


def test_sparseadapter_protocol(mocker: MockFixture) -> None:
    mock_call = mocker.patch.object(CustomSparseAdapter, "structure")
    mock_call2 = mocker.patch.object(CustomSparseAdapter, "read")
    mock_call3 = mocker.patch.object(CustomSparseAdapter, "read_block")
    mock_call4 = mocker.patch.object(CustomSparseAdapter, "specs")
    mock_call5 = mocker.patch.object(CustomSparseAdapter, "metadata")

    array = numpy.random.rand(2, 512, 512)

    structure = COOStructure(
        shape=(2 * 5,),
        chunks=((5, 5),),
        data_type=BuiltinDtype.from_numpy_dtype(array.dtype),
    )
    blocks: Dict[Tuple[int, ...], Tuple[NDArray[Any], Any]] = {(1,): (array, (1,))}
    metadata: JSON = {"foo": "bar"}
    anyslice = NDSlice(1, 1, 1)
    anyblock = NDSlice(1, 1, 1)

    anysparseadapter = CustomSparseAdapter(blocks, structure, metadata=metadata)
    assert anysparseadapter.structure_family == StructureFamily.sparse

    sparseadapter_protocol_functions(anysparseadapter, anyslice, anyblock)
    mock_call.assert_called_once()
    mock_call2.assert_called_once_with(anyslice)
    mock_call3.assert_called_once_with(anyblock)
    mock_call4.assert_called_once()
    mock_call5.assert_called_once()


class CustomTableAdapter:
    structure_family: Literal[StructureFamily.table] = StructureFamily.table
    supported_storage: Set[type[Storage]] = set()

    def __init__(
        self,
        partitions: List[Any],
        structure: TableStructure,
        *,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
    ) -> None:
        """

        Parameters
        ----------
        partitions :
        structure :
        metadata :
        specs :
        """
        self._metadata = metadata or {}
        self._partitions = list(partitions)
        self._structure = structure
        self._specs = specs or []

    def structure(self) -> TableStructure:
        return self._structure

    def read(
        self, fields: Optional[List[str]] = None
    ) -> Union[dask.dataframe.DataFrame, pandas.DataFrame]:
        return self._partitions

    def read_partition(
        self,
        partition: int,
        fields: Optional[str] = None,
    ) -> Union[dask.dataframe.DataFrame, pandas.DataFrame]:
        return self._partitions[partition]

    def specs(self) -> List[Spec]:
        return self._specs

    def metadata(self) -> JSON:
        return self._metadata

    def __getitem__(self, key: str) -> ArrayAdapter:
        array = numpy.random.rand(2, 512, 512)
        metadata: JSON = {"foo": "bar"}
        structure = ArrayStructure(
            data_type=BuiltinDtype.from_numpy_dtype(numpy.dtype("int32")),
            shape=(2, 512, 512),
            chunks=((1, 1), (512,), (512,)),
            dims=("time", "x", "y"),  # optional
        )
        return CustomArrayAdapter(array, structure, metadata=metadata)


def tableadapter_protocol_functions(
    adapter: TableAdapter, partition: int, fields: List[str]
) -> None:
    adapter.structure()
    adapter.read(fields)
    adapter.read_partition(partition)
    adapter.specs()
    adapter.metadata()
    adapter["abc"]


def test_tableadapter_protocol(mocker: MockFixture) -> None:
    mock_call = mocker.patch.object(CustomTableAdapter, "structure")
    mock_call2 = mocker.patch.object(CustomTableAdapter, "read")
    mock_call3 = mocker.patch.object(CustomTableAdapter, "read_partition")
    mock_call4 = mocker.patch.object(CustomTableAdapter, "specs")
    mock_call5 = mocker.patch.object(CustomTableAdapter, "metadata")
    mock_call6 = mocker.patch.object(CustomTableAdapter, "__getitem__")

    structure = TableStructure(
        arrow_schema="a", npartitions=1, columns=["A"], resizable=False
    )

    partitions = pandas.DataFrame([1])
    metadata: JSON = {"foo": "bar"}
    fields = ["a", "b", "c"]
    partition = 1

    anytableadapter = CustomTableAdapter(partitions, structure, metadata=metadata)
    assert anytableadapter.structure_family == StructureFamily.table

    tableadapter_protocol_functions(anytableadapter, partition, fields)
    mock_call.assert_called_once()
    mock_call2.assert_called_once_with(fields)
    mock_call3.assert_called_once_with(partition)
    mock_call4.assert_called_once()
    mock_call5.assert_called_once()
    mock_call6.assert_called_once_with("abc")


class CustomAccessPolicy(AccessPolicy):
    ALL = ALL_ACCESS

    def __init__(self, scopes: Optional[Scopes] = None) -> None:
        self.scopes = scopes if (scopes is not None) else ALL_SCOPES

    def _get_id(self, principal: Principal) -> None:
        return None

    async def allowed_scopes(
        self,
        node: BaseAdapter,
        principal: Principal,
        authn_access_tags: Optional[Set[str]],
        authn_scopes: Scopes,
    ) -> Scopes:
        allowed = self.scopes
        somemetadata = node.metadata()  # noqa: 841
        return allowed

    async def filters(
        self,
        node: BaseAdapter,
        principal: Principal,
        authn_access_tags: Optional[Set[str]],
        authn_scopes: Scopes,
        scopes: Scopes,
    ) -> Filters:
        queries: Filters = []
        somespecs = node.specs()  # noqa: 841
        return queries


async def accesspolicy_protocol_functions(
    policy: AccessPolicy,
    node: BaseAdapter,
    principal: Principal,
    authn_access_tags: Optional[Set[str]],
    authn_scopes: Scopes,
    scopes: Scopes,
) -> None:
    await policy.allowed_scopes(node, principal, authn_access_tags, authn_scopes)
    await policy.filters(node, principal, authn_access_tags, authn_scopes, scopes)


@pytest.mark.asyncio  # type: ignore
async def test_accesspolicy_protocol(mocker: MockFixture) -> None:
    mock_call = mocker.patch.object(CustomAwkwardAdapter, "metadata")
    mock_call2 = mocker.patch.object(CustomAwkwardAdapter, "specs")

    anyaccesspolicy = CustomAccessPolicy(scopes={"a12mdjnk4"})

    structure = AwkwardStructure(length=2, form={"a": "b"})

    metadata: JSON = {"foo": "bar"}
    container = DirectoryContainer(directory=Path("somedirectory"), form={})
    principal = Principal(
        uuid="12345678124123412345678123456781", type=PrincipalType.user
    )
    authn_access_tags = {"qux", "quux"}
    authn_scopes = {"abc", "baz"}
    scopes = {"abc"}

    anyawkwardadapter = CustomAwkwardAdapter(container, structure, metadata=metadata)

    await accesspolicy_protocol_functions(
        anyaccesspolicy,
        anyawkwardadapter,
        principal,
        authn_access_tags,
        authn_scopes,
        scopes,
    )
    mock_call.assert_called_once()
    mock_call2.assert_called_once()



import pytest

from tiled.catalog import in_memory
from tiled.client import Context, from_context, record_history
from tiled.server.app import build_app

N = 10  # number of items generated in sample


@pytest.fixture(scope="module")
def client(tmpdir_module):
    catalog = in_memory(writable_storage=[tmpdir_module])
    app = build_app(catalog)
    with Context.from_app(app) as context:
        client = from_context(context)
        for i in range(N):
            client.create_container(metadata={"num": i}, key=str(i))
        yield client


def test_first_value(client):
    "Fetching the first value requests a page of size 1."
    with record_history() as history:
        item = client.values().first()
    assert len(history.requests) == 1
    assert history.requests[0].url.params["page[limit]"] == "1"
    assert item.metadata["num"] == 0


def test_first_key(client):
    "Fetching the first key requests a page of size 1."
    with record_history() as history:
        key = client.keys().first()
    assert len(history.requests) == 1
    assert history.requests[0].url.params["page[limit]"] == "1"
    assert key == "0"


def test_last_value(client):
    "Fetching the last value requests a page of size 1."
    with record_history() as history:
        item = client.values().last()
    assert len(history.requests) == 1
    assert history.requests[0].url.params["page[limit]"] == "1"
    assert item.metadata["num"] == N - 1


def test_last_key(client):
    "Fetching the last key requests a page of size 1."
    with record_history() as history:
        key = client.keys().last()
    assert len(history.requests) == 1
    assert history.requests[0].url.params["page[limit]"] == "1"
    assert key == str(N - 1)


def test_head_values(client):
    "Fetching the 'head' values requests a page of size 5."
    with record_history() as history:
        items = client.values().head(5)
    assert len(history.requests) == 1
    assert history.requests[0].url.params["page[limit]"] == "5"
    actual_nums = [item.metadata["num"] for item in items]
    expected_nums = [0, 1, 2, 3, 4]
    assert actual_nums == expected_nums


def test_head_keys(client):
    "Fetching the 'head' keys requests a page of size 5."
    with record_history() as history:
        keys = client.keys().head(5)
    assert len(history.requests) == 1
    assert history.requests[0].url.params["page[limit]"] == "5"
    actual_nums = [int(key) for key in keys]
    expected_nums = [0, 1, 2, 3, 4]
    assert actual_nums == expected_nums


def test_tail_values(client):
    "Fetching the 'tail' values requests a page of size 5."
    with record_history() as history:
        items = client.values().tail(5)
    assert len(history.requests) == 1
    assert history.requests[0].url.params["page[limit]"] == "5"
    actual_nums = [item.metadata["num"] for item in items]
    expected_nums = [5, 6, 7, 8, 9]
    assert actual_nums == expected_nums


def test_tail_keys(client):
    "Fetching the 'tail' keys requests a page of size 5."
    with record_history() as history:
        keys = client.keys().tail(5)
    assert len(history.requests) == 1
    assert history.requests[0].url.params["page[limit]"] == "5"
    actual_nums = [int(key) for key in keys]
    expected_nums = [5, 6, 7, 8, 9]
    assert actual_nums == expected_nums


def test_middle_forward_values(client):
    "Fetching a slice of values in the middle requests a page of the correct size."
    with record_history() as history:
        items = client.values()[2:6]
    assert len(history.requests) == 1
    assert history.requests[0].url.params["page[limit]"] == "4"
    actual_nums = [item.metadata["num"] for item in items]
    expected_nums = [2, 3, 4, 5]
    assert actual_nums == expected_nums


def test_middle_forward_keys(client):
    "Fetching a slice of keys in the middle requests a page of the correct size."
    with record_history() as history:
        keys = client.keys()[2:6]
    assert len(history.requests) == 1
    assert history.requests[0].url.params["page[limit]"] == "4"
    actual_nums = [int(key) for key in keys]
    expected_nums = [2, 3, 4, 5]
    assert actual_nums == expected_nums


def test_middle_backward_values(client):
    "Fetching a slice of values in the middle requests a page of the correct size."
    with record_history() as history:
        items = client.values()[-2:-6:-1]
    assert len(history.requests) == 1
    assert history.requests[0].url.params["page[limit]"] == "4"
    actual_nums = [item.metadata["num"] for item in items]
    expected_nums = [8, 7, 6, 5]
    assert actual_nums == expected_nums


def test_middle_backward_keys(client):
    "Fetching a slice of keys in the middle requests a page of the correct size."
    with record_history() as history:
        keys = client.keys()[-2:-6:-1]
    assert len(history.requests) == 1
    assert history.requests[0].url.params["page[limit]"] == "4"
    actual_nums = [int(key) for key in keys]
    expected_nums = [8, 7, 6, 5]
    assert actual_nums == expected_nums


def test_manual_page_size_values(client):
    "The page_size method on values() can set the page size manually."
    with record_history() as history:
        items = client.values().page_size(2).head(5)
    assert len(history.requests) == 3
    for request in history.requests:
        assert request.url.params["page[limit]"] == "2"
    actual_nums = [item.metadata["num"] for item in items]
    expected_nums = [0, 1, 2, 3, 4]
    assert actual_nums == expected_nums


def test_manual_page_size_keys(client):
    "The page_size method on keys() can set the page size manually."
    with record_history() as history:
        keys = client.keys().page_size(2).head(5)
    assert len(history.requests) == 3
    for request in history.requests:
        assert request.url.params["page[limit]"] == "2"
    actual_nums = [int(key) for key in keys]
    expected_nums = [0, 1, 2, 3, 4]
    assert actual_nums == expected_nums


def test_manual_page_size_truncated_values(client):
    "If the manual page size is larger than the result set, it is truncated."
    with record_history() as history:
        items = client.values().page_size(6).head(5)
    assert len(history.requests) == 1
    assert history.requests[0].url.params["page[limit]"] == "5"
    actual_nums = [item.metadata["num"] for item in items]
    expected_nums = [0, 1, 2, 3, 4]
    assert actual_nums == expected_nums


def test_manual_page_size_truncated_keys(client):
    "If the manual page size is larger than the result set, it is truncated."
    with record_history() as history:
        keys = client.keys().page_size(6).head(5)
    assert len(history.requests) == 1
    assert history.requests[0].url.params["page[limit]"] == "5"
    actual_nums = [int(key) for key in keys]
    expected_nums = [0, 1, 2, 3, 4]
    assert actual_nums == expected_nums


def test_unbounded_values_slice(client):
    "An unbounded slice lets the server set the page size."
    with record_history() as history:
        items = client.values()[3:]
    assert len(history.requests) == 1
    assert "page[limit]" not in history.requests[0].url.params
    actual_nums = [item.metadata["num"] for item in items]
    expected_nums = [3, 4, 5, 6, 7, 8, 9]
    assert actual_nums == expected_nums


def test_unbounded_keys_slice(client):
    "An unbounded slice lets the server set the page size."
    with record_history() as history:
        keys = client.keys()[3:]
    assert len(history.requests) == 1
    assert "page[limit]" not in history.requests[0].url.params
    actual_nums = [int(key) for key in keys]
    expected_nums = [3, 4, 5, 6, 7, 8, 9]
    assert actual_nums == expected_nums



import sys

from ..structures.core import StructureFamily


def read(node, verbose=False, strict=False):
    """

    Parameters
    ----------
    node : tiled node
        A Client node with access to the tiled server.
        It can take form of anything compatible with the values in StructureFamily.
    verbose : bool, optional
        Enables status messages while process runs. The default is False.
    strict : bool, optional
        Enables debugging mode if a faulty node is found.

    Returns
    -------
    faulty_docs : List
        A list with URIs of all the faulty nodes that were found.

    """

    faulty_entries = []
    if node.structure_family == StructureFamily.container:
        for key, child_node in node.items():
            fault_result = read(child_node, verbose=verbose, strict=strict)
            faulty_entries.extend(fault_result)
    else:
        try:
            tmp = node.read()  # noqa: F841
        except Exception as err:
            faulty_entries.append(node.uri)
            if verbose:
                print(f"ERROR: {node.item['id']} - {err!r}", file=sys.stderr)
            if strict:
                raise
        else:
            if verbose:
                print(f"SUCCESS: {node.item['id']} ", file=sys.stderr)

    return faulty_entries



import platform
from pathlib import Path

import httpx
import pyarrow
import pytest

from tiled.client import from_uri
from tiled.server import SimpleTiledServer


def test_default():
    "Smoke test a server with defaults (no parameters)"
    with SimpleTiledServer() as server:
        client = from_uri(server.uri)

        # Write and read array data
        x = client.write_array([1, 2, 3], key="x")
        x[:]

        # Write and read tabular data to the SQL storage
        table = pyarrow.Table.from_pydict({"a": [1, 2, 3], "b": [4.0, 5.0, 6.0]})
        y = client.create_appendable_table(table.schema, key="y")
        y.append_partition(table, 0)
        y.read()

        repr(server)
        server._repr_html_()  # impl, used by Jupyter
        # Web UI
        response = httpx.get(server.web_ui_link).raise_for_status()
        assert response.headers["content-type"].startswith("text/html")


def test_one_at_a_time():
    "We cannot run two uvicorn servers in one process."
    # Two servers start on different ports.
    MSG = "Only one server can be run at a time in a given Python process."
    with SimpleTiledServer():
        with pytest.raises(RuntimeError, match=MSG):
            SimpleTiledServer()


def test_specified_port():
    "Run server on a user-specified port instead of a random one."
    ARBITRARY_PORT = 38593  # I hope it is free!
    with SimpleTiledServer(port=ARBITRARY_PORT) as server:
        assert server.port == ARBITRARY_PORT


def test_specified_api_key():
    "Run server with a user-specified API key instead of a random one."
    API_KEY = "secret"
    with SimpleTiledServer(api_key=API_KEY) as server:
        assert server.api_key == API_KEY


def test_persistent_data(tmp_path):
    "Write data in a specified location. Access it across a server restart."
    with SimpleTiledServer(directory=tmp_path) as server1:
        client1 = from_uri(server1.uri)
        client1.write_array([1, 2, 3], key="x")
        table = pyarrow.Table.from_pydict({"a": [1, 2, 3], "b": [4.0, 5.0, 6.0]})
        y = client1.create_appendable_table(table.schema, key="y")
        y.append_partition(table, 0)
        assert "x" in client1
        assert "y" in client1
    with SimpleTiledServer(directory=tmp_path) as server2:
        client2 = from_uri(server2.uri)
        assert "x" in client2
        assert "y" in client2
        assert client2["x"].read() is not None
        assert client2["y"].read() is not None
    assert server1.directory == server2.directory == tmp_path


def test_cleanup(tmp_path):
    if platform.system() == "Windows":
        # Windows cannot delete the logfiles because the global Python
        # logging system still has the logfiles open for appending.
        pytest.skip("Temp data directory is not cleaned up on Windows.")
    # Temp dir defined by SimpleTileServer is cleaned up.
    with SimpleTiledServer() as server:
        pass
    assert not Path(server.directory).exists()

    # Directory provided by user (which happens to be temp as well,
    # because this is a test) is _not_ cleaned up.
    with SimpleTiledServer(tmp_path) as server:
        pass
    assert Path(server.directory).exists()



import dataclasses
import functools
import os
from abc import abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Optional, Union
from urllib.parse import urlparse, urlunparse

import sqlalchemy.pool

if TYPE_CHECKING:
    import adbc_driver_manager.dbapi

from .utils import ensure_uri, path_from_uri, sanitize_uri

__all__ = [
    "EmbeddedSQLStorage",
    "RemoteSQLStorage",
    "FileStorage",
    "SQLStorage",
    "Storage",
    "get_storage",
    "parse_storage",
]


@dataclasses.dataclass(frozen=True)
class Storage:
    "Base class for representing storage location"
    uri: str

    def __post_init__(self):
        object.__setattr__(self, "uri", ensure_uri(self.uri))


@dataclasses.dataclass(frozen=True)
class FileStorage(Storage):
    "Filesystem storage location"

    @functools.cached_property
    def path(self):
        return path_from_uri(self.uri)


def _ensure_writable_location(uri: str) -> Path:
    "Ensure path is writable to avoid a confusing error message from driver."
    filepath = path_from_uri(uri)
    directory = Path(filepath).parent
    if directory.exists():
        if not os.access(directory, os.X_OK | os.W_OK):
            raise ValueError(
                f"The directory {directory} exists but is not writable and executable."
            )
        if Path(filepath).is_file() and (not os.access(filepath, os.W_OK)):
            raise ValueError(f"The path {filepath} exists but is not writable.")
    else:
        raise ValueError(f"The directory {directory} does not exist.")
    return filepath


@dataclasses.dataclass(frozen=True)
class SQLStorage(Storage):
    "General purpose SQL database storage with connection pooling"

    pool_size: int = 5
    max_overflow: int = 10

    def __post_init__(self):
        # Ensure pool_size and max_overflow are set to a default if not specified
        if self.dialect == "duckdb":
            # DuckDB does not support pooling, so we use StaticPool
            object.__setattr__(self, "pool_size", 1)
            object.__setattr__(self, "max_overflow", 0)
        else:
            if self.pool_size is None:
                object.__setattr__(self, "pool_size", 5)
            if self.max_overflow is None:
                object.__setattr__(self, "max_overflow", 10)

        super().__post_init__()

    @abstractmethod
    def create_adbc_connection(self) -> "adbc_driver_manager.dbapi.Connection":
        "Create a connection to the database."

        raise NotImplementedError("Subclasses must implement this method.")

    @property
    def dialect(self) -> str:
        "The database dialect (e.g. 'postgresql', 'sqlite', or 'duckdb')."
        return urlparse(self.uri).scheme

    @functools.cached_property
    def _adbc_connection(self) -> "adbc_driver_manager.dbapi.Connection":
        "A persistent connection to the database."
        return self.create_adbc_connection()

    @functools.cached_property
    def _connection_pool(self) -> "sqlalchemy.pool.QueuePool":
        creator = self._adbc_connection.adbc_clone
        if (self.dialect == "duckdb") or (":memory:" in self.uri):
            return sqlalchemy.pool.StaticPool(creator)
        else:
            return sqlalchemy.pool.QueuePool(
                creator, pool_size=self.pool_size, max_overflow=self.max_overflow
            )

    def connect(self) -> "adbc_driver_manager.dbapi.Connection":
        "Get a connection from the pool."
        return self._connection_pool.connect()

    def dispose(self) -> None:
        "Close all connections and dispose of the connection pool."
        self._connection_pool.dispose()
        self._adbc_connection.close()


@dataclasses.dataclass(frozen=True)
class EmbeddedSQLStorage(SQLStorage):
    "File-based SQL database storage location"

    def create_adbc_connection(self) -> "adbc_driver_manager.dbapi.Connection":
        filepath = _ensure_writable_location(self.uri)

        if self.uri.startswith("duckdb:"):
            import adbc_driver_duckdb.dbapi

            return adbc_driver_duckdb.dbapi.connect(str(filepath))

        elif self.uri.startswith("sqlite:"):
            import adbc_driver_sqlite.dbapi

            return adbc_driver_sqlite.dbapi.connect(str(filepath))

        else:
            raise ValueError(
                f"Unsupported URI scheme {self.uri}: use 'duckdb:' or 'sqlite:'"
            )


@dataclasses.dataclass(frozen=True)
class RemoteSQLStorage(SQLStorage):
    "Authenticated server-based SQL database storage location"
    username: Optional[str] = None
    password: Optional[str] = None

    def __post_init__(self):
        # Ensure the URI is sanitized and credentials are stored separately
        if not self.uri.startswith("postgresql:"):
            raise ValueError(f"Unsupported URI scheme {self.uri}: use 'postgresql:'")

        uri, username, password = sanitize_uri(self.uri)
        if (username is not None) or (password is not None):
            if (self.username is not None) or (self.password is not None):
                raise ValueError(
                    "Credentials passed both in URI and in username/password fields."
                )
            object.__setattr__(self, "uri", uri)
            object.__setattr__(self, "username", username)
            object.__setattr__(self, "password", password)

        super().__post_init__()

    def __repr__(self):
        return f"SQLStorage(uri={self.uri!r}, username={self.username!r})"

    @functools.cached_property
    def authenticated_uri(self):
        parsed_uri = urlparse(self.uri)
        components = (
            parsed_uri.scheme,
            f"{self.username}:{self.password}@{parsed_uri.netloc}",
            parsed_uri.path,
            parsed_uri.params,
            parsed_uri.query,
            parsed_uri.fragment,
        )

        return urlunparse(components)

    def create_adbc_connection(self) -> "adbc_driver_manager.dbapi.Connection":
        import adbc_driver_postgresql.dbapi

        return adbc_driver_postgresql.dbapi.connect(self.authenticated_uri)


def parse_storage(
    item: Union[Path, str],
    *,
    pool_size: Optional[int] = None,
    max_overflow: Optional[int] = None,
) -> Storage:
    "Create a Storage object from a URI or Path."
    item = ensure_uri(item)
    scheme = urlparse(item).scheme
    if scheme == "file":
        result = FileStorage(item)
    elif scheme == "postgresql":
        result = RemoteSQLStorage(item, pool_size=pool_size, max_overflow=max_overflow)
    elif scheme in {"sqlite", "duckdb"}:
        result = EmbeddedSQLStorage(
            item, pool_size=pool_size, max_overflow=max_overflow
        )
    else:
        raise ValueError(f"writable_storage item {item} has unrecognized scheme")
    return result


# This global registry enables looking up Storage via URI, primarily for the
# purpose of obtaining credentials, which are not stored in the catalog
# database.
_STORAGE: Dict[str, Storage] = {}


def register_storage(storage: Storage) -> None:
    "Stash Storage for later lookup by URI."
    _STORAGE[storage.uri] = storage


def get_storage(uri: str) -> Storage:
    "Look up Storage by URI."
    return _STORAGE[uri]



# This is a variation on starlette's FileResponse that adds support for the
# 'Range' HTTP header.

# It is adapted from a closed PR in starlette which was reviewed by a core
# starlette maintainer but put aside for now in favor of other priorities in
# starlette development. Thus, we implement it here in tiled. If in the future
# starlette adds support upstream, we should consider refactoring to use that.

# Ref: https://github.com/encode/starlette/pull/1999
import hashlib
import os
import stat
import typing

import anyio
from starlette.responses import FileResponse, Receive, Scope, Send, formatdate
from starlette.status import HTTP_200_OK, HTTP_206_PARTIAL_CONTENT


class FileResponseWithRange(FileResponse):
    def __init__(
        self,
        path: typing.Union[str, "os.PathLike[str]"],
        status_code: int = HTTP_200_OK,
        *args,
        range: typing.Optional[typing.Tuple[int, int]] = None,
        **kwargs,
    ):
        if (range is not None) and (status_code != HTTP_206_PARTIAL_CONTENT):
            raise RuntimeError(
                f"Range requests must have a {HTTP_206_PARTIAL_CONTENT} status code."
            )
        self.range = range
        super().__init__(path, status_code, *args, **kwargs)

    def set_stat_headers(self, stat_result: os.stat_result) -> None:
        content_length = str(stat_result.st_size)
        size = str(stat_result.st_size)
        last_modified = formatdate(stat_result.st_mtime, usegmt=True)
        etag_base = str(stat_result.st_mtime) + "-" + str(stat_result.st_size)
        if self.range is not None:
            start, end = self.range
            etag_base += f"-{start}/{end}"
            content_length = str(end - start + 1)
            self.headers.setdefault("accept-ranges", "bytes")
            self.headers.setdefault("content-range", f"bytes {start}-{end}/{size}")
        else:
            content_length = size
        etag = hashlib.md5(etag_base.encode(), usedforsecurity=False).hexdigest()

        self.headers.setdefault("content-length", content_length)
        self.headers.setdefault("last-modified", last_modified)
        self.headers.setdefault("etag", etag)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if self.stat_result is None:
            try:
                stat_result = await anyio.to_thread.run_sync(os.stat, self.path)
                self.set_stat_headers(stat_result)
            except FileNotFoundError:
                raise RuntimeError(f"File at path {self.path} does not exist.")
            else:
                mode = stat_result.st_mode
                if not stat.S_ISREG(mode):
                    raise RuntimeError(f"File at path {self.path} is not a file.")
        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self.raw_headers,
            }
        )
        if scope["method"].upper() == "HEAD":
            await send({"type": "http.response.body", "body": b"", "more_body": False})
        elif "extensions" in scope and "http.response.pathsend" in scope["extensions"]:
            await send({"type": "http.response.pathsend", "path": str(self.path)})
        else:
            async with await anyio.open_file(self.path, mode="rb") as file:
                if self.range is not None:
                    start, end = self.range
                    await file.seek(start)
                else:
                    start, end = 0, stat_result.st_size - 1
                remaining_bytes = end - start + 1
                more_body = True
                while more_body:
                    chunk_size = min(remaining_bytes, self.chunk_size)
                    chunk = await file.read(chunk_size)
                    remaining_bytes -= len(chunk)
                    more_body = remaining_bytes > 0 and len(chunk) == chunk_size
                    await send(
                        {
                            "type": "http.response.body",
                            "body": chunk,
                            "more_body": more_body,
                        }
                    )
        if self.background is not None:
            await self.background()



import os

import numpy
import pytest

from ..adapters import hdf5 as hdf5_adapters
from ..adapters.hdf5 import HDF5Adapter
from ..adapters.mapping import MapAdapter
from ..catalog import in_memory
from ..client import Context, from_context, record_history
from ..server.app import build_app
from ..structures.core import StructureFamily
from ..structures.data_source import Asset, DataSource, Management
from ..utils import BrokenLink, ensure_uri, path_from_uri
from ..utils import tree as tree_util


@pytest.fixture(scope="module")
def tree(tmp_path_factory):
    return in_memory(writable_storage=tmp_path_factory.getbasetemp())


@pytest.fixture(scope="module")
def context(tree):
    with Context.from_app(build_app(tree)) as context:
        yield context


@pytest.fixture(scope="module")
def example_file(tmp_path_factory):
    h5py = pytest.importorskip("h5py")
    file_path = tmp_path_factory.mktemp("data").joinpath("example.h5")
    with h5py.File(file_path, "w") as file:
        a = file.create_group("a")
        b = a.create_group("b")
        c = b.create_group("c")
        c.create_dataset("d", data=numpy.ones((3, 3)))

    yield ensure_uri(file_path)


@pytest.fixture(scope="module")
def example_file_with_empty_data(tmp_path_factory):
    h5py = pytest.importorskip("h5py")
    file_path = tmp_path_factory.mktemp("data").joinpath("example_with_empty.h5")
    with h5py.File(file_path, "w") as file:
        a = file.create_group("a")
        b = a.create_group("b")
        c = b.create_group("c")
        c.create_dataset("d", data=numpy.empty(shape=0))
        c.create_dataset("e", data=numpy.empty(shape=(5, 7)))
        c.create_dataset("f", data=numpy.empty(shape=1, dtype="S0"))
        c.create_dataset("g", data=[])
        c.create_dataset("h", data="")
        c.create_dataset("i", data=numpy.empty(shape=()))

    yield ensure_uri(file_path)


@pytest.fixture(scope="module")
def example_file_with_scalars(tmp_path_factory):
    h5py = pytest.importorskip("h5py")
    file_path = tmp_path_factory.mktemp("data").joinpath("example_with_scalars.h5")
    with h5py.File(file_path, "w") as file:
        a = file.create_group("a")
        b = a.create_group("b")
        c = b.create_group("c")
        c.create_dataset("int", data=42)
        c.create_dataset("float", data=3.14)
        c.create_dataset("str", data="hello")
        c.create_dataset("bytes", data=b"hello")
        c.create_dataset("bool", data=True)

    yield ensure_uri(file_path)


@pytest.fixture(scope="module")
def example_file_with_vlen_str_in_dataset(tmp_path_factory):
    h5py = pytest.importorskip("h5py")
    file_path = tmp_path_factory.mktemp("data").joinpath("example_with_vlen_str.h5")
    with h5py.File(file_path, "w") as file:
        a = file.create_group("a")
        b = a.create_group("b")
        c = b.create_group("c")
        # Need to do this to make a vlen str dataset
        dt = h5py.string_dtype(encoding="utf-8")
        dset = c.create_dataset("d", (100,), dtype=dt)
        assert dset.dtype == "object"
        dset[0] = b"test"

    yield ensure_uri(file_path)


@pytest.fixture(scope="function")
def example_file_with_links(tmp_path_factory):
    h5py = pytest.importorskip("h5py")
    file_path = tmp_path_factory.mktemp("data").joinpath("example.h5")
    with h5py.File(file_path.with_name("linked.h5"), "w") as file:
        z = file.create_group("z")
        y = z.create_group("y")
        y.create_dataset("x", data=2 * numpy.ones((5, 5)))
    with h5py.File(file_path, "w") as file:
        a = file.create_group("a")
        b = a.create_group("b")
        c = b.create_group("c")
        c.create_dataset("d", data=numpy.ones((3, 3)))
        b["hard_link"] = c["d"]
        b["soft_link"] = h5py.SoftLink("/a/b/c/d")
        b["extr_link"] = h5py.ExternalLink("linked.h5", "/z/y")

    yield ensure_uri(file_path)


def test_from_file(example_file, buffer):
    """Serve a single HDF5 file at top level."""
    h5py = pytest.importorskip("h5py")
    tree = HDF5Adapter.from_uris(example_file)
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
        arr = client["a"]["b"]["c"]["d"].read()
        assert isinstance(arr, numpy.ndarray)
        client.export(buffer, format="application/x-hdf5")
        file = h5py.File(buffer, "r")
        assert file["a"]["b"]["c"]["d"] is not None


@pytest.mark.filterwarnings("ignore: The dataset")
@pytest.mark.parametrize("key", ["d", "e", "f", "g", "h", "i"])
def test_from_file_with_empty_data(example_file_with_empty_data, buffer, key):
    """Serve a single HDF5 file at top level."""
    h5py = pytest.importorskip("h5py")
    tree = HDF5Adapter.from_uris(example_file_with_empty_data)
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
        arr = client["a"]["b"]["c"][key].read()
        assert isinstance(arr, numpy.ndarray)
        client.export(buffer, format="application/x-hdf5")
        file = h5py.File(buffer, "r")
        assert file["a"]["b"]["c"][key] is not None


@pytest.mark.filterwarnings("ignore: The dataset")
@pytest.mark.parametrize("key", ["bool", "int", "float", "str", "bytes"])
@pytest.mark.parametrize("num", [1, 5])
def test_from_file_with_scalars(example_file_with_scalars, buffer, key: str, num: int):
    """Serve HDF5 file(s) containing scalars."""
    if key in {"str", "bytes"}:
        pytest.xfail("HDF5Adapter treats object dtypes differently.")
    h5py = pytest.importorskip("h5py")
    tree = HDF5Adapter.from_uris(*[example_file_with_scalars] * num)
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
        arr = client["a"]["b"]["c"][key].read()
        assert isinstance(arr, numpy.ndarray)
        assert arr.shape == (num,)
        client.export(buffer, format="application/x-hdf5")
        file = h5py.File(buffer, "r")
        assert file["a"]["b"]["c"][key] is not None


@pytest.mark.filterwarnings("ignore: The dataset")
def test_from_file_with_vlen_str_dataset(example_file_with_vlen_str_in_dataset, buffer):
    """Serve a single HDF5 file at top level."""
    h5py = pytest.importorskip("h5py")
    tree = HDF5Adapter.from_uris(example_file_with_vlen_str_in_dataset)
    with pytest.warns(UserWarning):
        with Context.from_app(build_app(tree)) as context:
            client = from_context(context)
    arr = client["a"]["b"]["c"]["d"].read()
    assert isinstance(arr, numpy.ndarray)
    with pytest.warns(UserWarning):
        client.export(buffer, format="application/x-hdf5")
    file = h5py.File(buffer, "r")
    assert file["a"]["b"]["c"]["d"] is not None


def test_from_group(example_file, buffer):
    """Serve a Group within an HDF5 file."""
    h5py = pytest.importorskip("h5py")
    tree = HDF5Adapter.from_uris(example_file, dataset="a/b")
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
    arr = client["c"]["d"].read()
    assert isinstance(arr, numpy.ndarray)
    client.export(buffer, format="application/x-hdf5")
    file = h5py.File(buffer, "r")
    assert file["c"]["d"] is not None


def test_from_multiple(example_file, buffer):
    """Serve two files within a mapping."""
    h5py = pytest.importorskip("h5py")
    tree = MapAdapter(
        {
            "A": HDF5Adapter.from_uris(example_file),
            "B": HDF5Adapter.from_uris(example_file),
        }
    )
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
    arr_A = client["A"]["a"]["b"]["c"]["d"].read()
    assert isinstance(arr_A, numpy.ndarray)
    arr_B = client["B"]["a"]["b"]["c"]["d"].read()
    assert isinstance(arr_B, numpy.ndarray)
    client.export(buffer, format="application/x-hdf5")
    file = h5py.File(buffer, "r")
    assert file["A"]["a"]["b"]["c"]["d"] is not None
    assert file["B"]["a"]["b"]["c"]["d"] is not None


def test_inlined_contents(example_file):
    """Test that the recursive structure and metadata are inlined into one request."""
    tree = HDF5Adapter.from_uris(example_file)
    assert hdf5_adapters.INLINED_DEPTH > 1
    original = hdf5_adapters.INLINED_DEPTH
    try:
        with Context.from_app(build_app(tree)) as context:
            with record_history() as hN:
                client = from_context(context)
                tree_util(client)
            hdf5_adapters.INLINED_DEPTH = 1
            with record_history() as h1:
                client = from_context(context)
                tree_util(client)
            hdf5_adapters.INLINED_DEPTH = 0
            with record_history() as h0:
                client = from_context(context)
                tree_util(client)
            assert len(h0.requests) > len(h1.requests) > len(hN.requests)
    finally:
        hdf5_adapters.INLINED_DEPTH = original


def test_file_with_links(example_file_with_links, buffer):
    """Serve an HDF5 file with internal and external links."""

    h5py = pytest.importorskip("h5py")
    tree = HDF5Adapter.from_uris(example_file_with_links)
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)

    # Read the original array
    arr = client["a/b/c/d"].read()
    arr = client["a"]["b/c/d"].read()
    arr = client["a"]["b"]["c/d"].read()
    arr = client["a"]["b"]["c"]["d"].read()
    arr = client["a/b/c"]["d"].read()
    assert isinstance(arr, numpy.ndarray)
    assert numpy.allclose(arr, numpy.ones((3, 3)))

    # Read the hard link
    arr = client["a"]["b"]["hard_link"].read()
    assert isinstance(arr, numpy.ndarray)
    assert numpy.allclose(arr, numpy.ones((3, 3)))

    # Read the soft link
    arr = client["a"]["b"]["soft_link"].read()
    assert isinstance(arr, numpy.ndarray)
    assert numpy.allclose(arr, numpy.ones((3, 3)))

    # Read the external link
    arr = client["a"]["b"]["extr_link/x"].read()
    assert isinstance(arr, numpy.ndarray)
    assert numpy.allclose(arr, 2 * numpy.ones((5, 5)))

    # Export the tree into another file/buffer
    client.export(buffer, format="application/x-hdf5")
    file = h5py.File(buffer, "r")
    assert numpy.allclose(numpy.array(file["a/b/c/d"]), numpy.ones((3, 3)))
    assert numpy.allclose(numpy.array(file["a/b/hard_link"]), numpy.ones((3, 3)))
    assert numpy.allclose(numpy.array(file["a/b/soft_link"]), numpy.ones((3, 3)))
    assert numpy.allclose(numpy.array(file["a/b/extr_link/x"]), 2 * numpy.ones((5, 5)))


def test_file_with_broken_links(example_file_with_links):
    """Raise an error when accessing non-existing keys."""

    h5py = pytest.importorskip("h5py")
    main_file_path = path_from_uri(example_file_with_links)
    child_file_path = main_file_path.with_name("linked.h5")

    # Case 1. Broken soft link
    # KeyError: 'Unable to synchronously open object (component not found)'
    with h5py.File(main_file_path, "r+") as file:
        file["a/b/c"].pop("d")

    with pytest.raises(BrokenLink):
        tree = HDF5Adapter.from_uris(example_file_with_links, dataset="a/b/soft_link")

    with pytest.raises(KeyError):
        tree = HDF5Adapter.from_uris(example_file_with_links, dataset="a/b/c/d")

    tree = HDF5Adapter.from_uris(example_file_with_links, dataset="a")
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
    assert len(client["b"]) == 4  # All keys are there
    with pytest.raises(KeyError):
        client["b/soft_link"]

    # Case 2. Broken children of an external link
    # KeyError when accessing 'x'; 'y' is still accessible (but empty)
    with h5py.File(child_file_path, "r+") as file:
        file["z/y"].pop("x")

    with pytest.raises(KeyError):
        tree = HDF5Adapter.from_uris(example_file_with_links, dataset="a/b/extr_link/x")

    tree = HDF5Adapter.from_uris(example_file_with_links, dataset="a/b/extr_link")
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
    assert len(client) == 0
    with pytest.raises(KeyError):
        client["x"]

    # Case 3. Broken external link -- the referenced object is missing
    # KeyError: "Unable to synchronously open object (object 'y' doesn't exist)"
    with h5py.File(child_file_path, "r+") as file:
        file["z"].pop("y")

    with pytest.raises(KeyError):
        tree = HDF5Adapter.from_uris(example_file_with_links, dataset="a/b/extr_link")

    # A client can still be instantiated
    tree = HDF5Adapter.from_uris(example_file_with_links)
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)

    # Case 4. Broken external link -- the file is missing
    # KeyError: "Unable to synchronously open object (unable to open external file, external link file name = 'linked.h5')"  # noqa
    os.remove(child_file_path)

    with pytest.raises(BrokenLink):
        tree = HDF5Adapter.from_uris(example_file_with_links, dataset="a/b/extr_link")

    tree = HDF5Adapter.from_uris(example_file_with_links)
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
    with pytest.raises(KeyError):
        client["a/b/extr_link"]


def test_register_broken_hdf5_file(context, example_file_with_links):
    """Test that a broken HDF5 file can be registered and accessed."""

    client = from_context(context)

    h5py = pytest.importorskip("h5py")
    main_file_path = path_from_uri(example_file_with_links)
    child_file_path = main_file_path.with_name("linked.h5")

    # Brake the soft link
    with h5py.File(main_file_path, "r+") as file:
        file["a/b/c"].pop("d")

    # Brake the external link
    os.remove(child_file_path)

    asset = Asset(
        data_uri=f"file://localhost/{main_file_path}",
        is_directory=False,
        parameter="data_uris",
        num=0,
    )

    data_source_from_root = DataSource(
        mimetype="application/x-hdf5",
        assets=[asset],
        structure_family=StructureFamily.container,
        structure=None,
        management=Management.external,
    )

    data_source_from_node = DataSource(
        mimetype="application/x-hdf5",
        assets=[asset],
        structure_family=StructureFamily.container,
        structure=None,
        parameters={"dataset": "a/b"},
        management=Management.external,
    )

    data_source_from_soft = DataSource(
        mimetype="application/x-hdf5",
        assets=[asset],
        structure_family=StructureFamily.container,
        structure=None,
        parameters={"dataset": "a/b/soft_link"},
        management=Management.external,
    )

    data_source_from_extr = DataSource(
        mimetype="application/x-hdf5",
        assets=[asset],
        structure_family=StructureFamily.container,
        structure=None,
        parameters={"dataset": "a/b/extr_link"},
        management=Management.external,
    )

    client.new(
        structure_family=StructureFamily.container,
        data_sources=[data_source_from_root],
        key="ds_from_root",
    )

    client.new(
        structure_family=StructureFamily.container,
        data_sources=[data_source_from_node],
        key="ds_from_node",
    )

    client.new(
        structure_family=StructureFamily.container,
        data_sources=[data_source_from_soft],
        key="ds_from_soft",
    )

    client.new(
        structure_family=StructureFamily.container,
        data_sources=[data_source_from_extr],
        key="ds_from_extr",
    )

    assert len(client) == 4  # All registered
    assert set(client.keys()) == {
        "ds_from_root",
        "ds_from_node",
        "ds_from_soft",
        "ds_from_extr",
    }
    assert list(client["ds_from_root"].keys()) is not None
    assert list(client["ds_from_node"].keys()) is not None

    # Datasets referenced from the root of hdf5 file
    with pytest.raises(KeyError):
        client["ds_from_root"]["a/b/soft_link"]

    with pytest.raises(KeyError):
        client["ds_from_root"]["a/b/extr_link"]

    # Datasets referenced from an internal node of hdf5 file
    with pytest.raises(KeyError):
        client["ds_from_node"]["soft_link"]

    with pytest.raises(KeyError):
        client["ds_from_node"]["extr_link"]

    # Datasets referenced from links directly
    with pytest.raises(KeyError):
        list(client["ds_from_soft"].keys())

    with pytest.raises(KeyError):
        list(client["ds_from_extr"].keys())



import contextlib
from typing import Any, List, Optional, Set, Tuple

import dask.array
import numpy
import pandas
from numpy.typing import NDArray

from ..ndslice import NDSlice
from ..storage import Storage
from ..structures.array import ArrayStructure
from ..structures.core import Spec, StructureFamily
from ..type_aliases import JSON


class ArrayAdapter:
    """
    Wrap an array-like object in an interface that Tiled can serve.

    Examples
    --------

    Wrap any array-like.

    >>> ArrayAdapter.from_array(numpy.random.random((100, 100)))

    >>> ArrayAdapter.from_array(dask.array.from_array(numpy.random.random((100, 100)), chunks=(100, 50)))

    """

    structure_family = StructureFamily.array
    supported_storage: Set[type[Storage]] = set()

    def __init__(
        self,
        array: NDArray[Any],
        structure: ArrayStructure,
        *,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
    ) -> None:
        """

        Parameters
        ----------
        array :
        structure :
        metadata :
        specs :
        """
        self._array = array
        self._structure = structure
        self._metadata = metadata or {}
        self.specs = specs or []

    @classmethod
    def from_array(
        cls,
        array: NDArray[Any],
        *,
        shape: Optional[Tuple[int, ...]] = None,
        chunks: Optional[Tuple[Tuple[int, ...], ...]] = None,
        dims: Optional[Tuple[str, ...]] = None,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
    ) -> "ArrayAdapter":
        """

        Parameters
        ----------
        array :
        shape :
        chunks :
        dims :
        metadata :
        specs :

        Returns
        -------

        """
        # May be a list of something; convert to array
        if not hasattr(array, "__array__"):
            array = numpy.asanyarray(array)

        # Convert array of arrays to ND array to expose the underlying dtype
        is_array_of_arrays = (
            array.dtype == "object"
            and array.shape[0]
            and isinstance(array[0], numpy.ndarray)
        )
        if is_array_of_arrays:
            with contextlib.suppress(ValueError):
                # only uniform arrays (with same dimensions) are stackable
                array = numpy.vstack(array)

        # Convert (experimental) pandas.StringDtype to numpy's unicode string dtype
        is_likely_string_dtype = isinstance(array.dtype, pandas.StringDtype) or (
            array.dtype == "object" and array.dtype.fields is None
        )
        if is_likely_string_dtype:
            array = numpy.array([str(x) for x in array])  # becomes "<Un" dtype

        structure = ArrayStructure.from_array(
            array, shape=shape, chunks=chunks, dims=dims
        )
        return cls(
            array,
            structure=structure,
            metadata=metadata,
            specs=specs,
        )

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._array!r})"

    @property
    def dims(self) -> Optional[Tuple[str, ...]]:
        return self._structure.dims

    def metadata(self) -> JSON:
        return self._metadata

    def structure(self) -> ArrayStructure:
        return self._structure

    def read(
        self,
        slice: NDSlice = NDSlice(...),
    ) -> NDArray[Any]:
        """

        Parameters
        ----------
        slice :

        Returns
        -------

        """
        # _array[...] requires an actual tuple, not just a subclass of tuple
        array = self._array[tuple(slice)] if slice else self._array
        if isinstance(self._array, dask.array.Array):
            return array.compute()
        return array

    def read_block(
        self,
        block: Tuple[int, ...],
        slice: NDSlice = NDSlice(...),
    ) -> NDArray[Any]:
        """

        Parameters
        ----------
        block :
        slice :

        Returns
        -------

        """
        # Slice the whole array to get this block.
        slice_, _ = slice_and_shape_from_block_and_chunks(block, self._structure.chunks)
        # _array[...] requires an actual tuple, not just a subclass of tuple
        array = self._array[tuple(slice_)]
        # Slice within the block.
        array = array[slice] if slice else array
        if isinstance(self._array, dask.array.Array):
            return array.compute()
        return array


def slice_and_shape_from_block_and_chunks(
    block: Tuple[int, ...], chunks: Tuple[Tuple[int, ...], ...]
) -> tuple[NDSlice, NDSlice]:
    """
    Given dask-like chunks and block id, return slice and shape of the block.
    Parameters
    ----------
    block :
    chunks :

    Returns
    -------

    """
    slice_ = []
    shape = []
    for b, c in zip(block, chunks):
        start = sum(c[:b])
        dim = c[b]
        slice_.append(slice(start, start + dim))
        shape.append(dim)
    return NDSlice(*slice_), NDSlice(*shape)



import builtins
from typing import Any, List, Optional, Tuple, Union

import numpy
from numpy._typing import NDArray

from ..catalog.orm import Node
from ..ndslice import NDSlice
from ..structures.array import ArrayStructure, BuiltinDtype
from ..structures.core import Spec, StructureFamily
from ..structures.data_source import DataSource
from ..type_aliases import JSON
from ..utils import path_from_uri
from .resource_cache import with_resource_cache
from .sequence import FileSequenceAdapter
from .utils import init_adapter_from_catalog


class NPYAdapter:
    """
    Read the Numpy on-disk format, NPY (.npy).

    Examples
    --------

    >>> NPYAdapter("path/to/file.npy")
    """

    structure_family = StructureFamily.array

    def __init__(
        self,
        data_uri: str,
        *,
        structure: ArrayStructure,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
    ) -> None:
        """

        Parameters
        ----------
        data_uri :
        structure :
        metadata :
        specs :
        """
        self._filepath = path_from_uri(data_uri)
        self.specs = specs or []
        self._provided_metadata = metadata or {}
        self._structure = structure

    @classmethod
    def from_catalog(
        cls,
        data_source: DataSource[ArrayStructure],
        node: Node,
        /,
        **kwargs: Optional[Any],
    ) -> "NPYAdapter":
        return init_adapter_from_catalog(cls, data_source, node, **kwargs)  # type: ignore

    @classmethod
    def from_uris(
        cls,
        data_uri: str,
        **kwargs: Optional[Any],
    ) -> "NPYAdapter":
        filepath = path_from_uri(data_uri)
        cache_key = (numpy.load, filepath)
        arr = with_resource_cache(cache_key, numpy.load, filepath)

        structure = ArrayStructure(
            shape=arr.shape,
            chunks=tuple((dim,) for dim in arr.shape),
            data_type=BuiltinDtype.from_numpy_dtype(arr.dtype),
        )

        return cls(
            data_uri,
            structure=structure,
        )

    def metadata(self) -> JSON:
        return self._provided_metadata.copy()

    def read(self, slice: NDSlice = NDSlice(...)) -> NDArray[Any]:
        cache_key = (numpy.load, self._filepath)
        arr = with_resource_cache(cache_key, numpy.load, self._filepath)
        arr = arr[slice] if slice else arr
        return arr

    def read_block(
        self, block: Tuple[int, ...], slice: Optional[builtins.slice] = None
    ) -> NDArray[Any]:
        if sum(block) != 0:
            raise IndexError(block)
        cache_key = (numpy.load, self._filepath)
        arr = with_resource_cache(cache_key, numpy.load, self._filepath)
        arr = arr[slice] if slice else arr
        if slice is not None:
            arr = arr[slice]
        return arr

    def structure(self) -> ArrayStructure:
        return self._structure


class NPYSequenceAdapter(FileSequenceAdapter):
    def _load_from_files(
        self, slice: Union[builtins.slice, int] = slice(None)
    ) -> NDArray[Any]:
        if isinstance(slice, int):
            return numpy.load(self.filepaths[slice])[None, ...]
        else:
            return numpy.asarray([numpy.load(file) for file in self.filepaths[slice]])



import contextlib
import os
import re
import subprocess
import sys
import threading
from queue import Queue

import httpx
import pytest


@contextlib.contextmanager
def run_cli(command):
    "Run '/path/to/this/python -m ...'"
    process = subprocess.Popen(
        [sys.executable, "-m"] + command.split(),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    yield process
    process.terminate()


def scrape_server_url_from_logs(process):
    "Scrape from server logs 'Uvicorn running on https://...'"

    def target(queue):
        pattern = re.compile(r"Uvicorn running on .*(http:\/\/\S+:\d+).*")
        lines = []
        while not process.poll():
            line = process.stderr.readline()
            lines.append(line.decode())
            if match := pattern.search(line.decode()):
                break
        else:
            raise RuntimeError(
                "Did not find server URL in log output:\n" + "\n".join(lines)
            )
        url = match.group(1)
        queue.put(url)

    queue = Queue()
    thread = threading.Thread(target=target, args=(queue,))
    thread.start()
    url = queue.get(timeout=10)
    # If the server has an error starting up, the target() will
    # never find a match, and a TimeoutError will be raised above.
    # The thread will leak. This is the best reasonably simple,
    # portable approach available.
    thread.join()
    _, port = url.rsplit(":", 1)
    assert port != "8000"  # should be a random high port
    return url


def check_server_readiness(process):
    "Given a server process, check that it responds successfully to HTTP."
    url = scrape_server_url_from_logs(process)
    httpx.get(url).raise_for_status()


@pytest.mark.parametrize(
    "args",
    [
        "",
        "--verbose",
        "--api-key secret",
    ],
)
def test_serve_directory(args, tmp_path):
    "Test 'tiled serve directory ... with a variety of arguments."
    with run_cli(f"tiled serve directory {tmp_path!s} --port 0 " + args) as process:
        check_server_readiness(process)


@pytest.mark.parametrize(
    "args",
    [
        "",
        "--api-key secret",
    ],
)
def test_serve_catalog_temp(args, tmp_path):
    "Test 'tiled serve catalog --temp ... with a variety of arguments."
    with run_cli(f"tiled serve directory {tmp_path!s} --port 0 " + args) as process:
        check_server_readiness(process)


@pytest.mark.parametrize(
    "args",
    [
        "",
    ],
)
def test_serve_config(args, tmp_path, sqlite_or_postgres_uri):
    "Test 'tiled serve config' with a tmp config file."
    (tmp_path / "data").mkdir()
    (tmp_path / "config").mkdir()
    config_filepath = tmp_path / "config" / "config.yml"
    with open(config_filepath, "w") as file:
        file.write(
            f"""
authentication:
  allow_anonymous_access: false
trees:
  - path: /
    tree: catalog
    args:
      uri: {sqlite_or_postgres_uri}
      writable_storage: {tmp_path / 'data'}
      init_if_not_exists: true
"""
        )
    with run_cli(f"tiled serve config {config_filepath} --port 0 " + args) as process:
        check_server_readiness(process)


def test_cli_version():
    from tiled import __version__

    with run_cli("tiled --version") as process:
        assert process.stdout is not None
        line = process.stdout.readline()
    assert line.decode() == f"{__version__}{os.linesep}"



from ..utils import tree
from .constructors import from_context, from_profile, from_uri
from .container import ASCENDING, DESCENDING
from .context import Context
from .logger import hide_logs, record_history, show_logs
from .metadata_update import DELETE_KEY

__all__ = [
    "ASCENDING",
    "Context",
    "DESCENDING",
    "DELETE_KEY",
    "from_context",
    "from_profile",
    "from_uri",
    "hide_logs",
    "record_history",
    "show_logs",
    "tree",
]



import hashlib

import canonicaljson


def compute_structure_id(structure):
    "Compute HEX digest of MD5 hash of RFC 8785 canonical form of JSON."
    canonical_structure = canonicaljson.encode_canonical_json(structure)

    return hashlib.md5(canonical_structure).hexdigest()



import copy
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, Iterator, List, Optional, Tuple, Union
from urllib.parse import quote_plus

import dask.dataframe
import pandas

from ..catalog.orm import Node
from ..storage import FileStorage, Storage
from ..structures.array import ArrayStructure
from ..structures.core import Spec, StructureFamily
from ..structures.data_source import Asset, DataSource, Management
from ..structures.table import TableStructure
from ..type_aliases import JSON
from ..utils import ensure_uri, path_from_uri
from .array import ArrayAdapter
from .utils import init_adapter_from_catalog


class CSVAdapter:
    """Adapter for tabular data stored as partitioned text (csv) files"""

    structure_family = StructureFamily.table
    supported_storage = {FileStorage}

    def __init__(
        self,
        data_uris: Iterable[str],
        structure: Optional[TableStructure] = None,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
        **kwargs: Optional[Any],
    ) -> None:
        """Adapter for partitioned tabular data stored as a sequence of text (csv) files

        Parameters
        ----------
        data_uris : list of uris to csv files
        structure :
        metadata :
        specs :
        kwargs : dict
            any keyword arguments that can be passed to the pandas.read_csv function, e.g. names, sep, dtype, etc.
        """
        self._file_paths = [path_from_uri(uri) for uri in data_uris]
        self._metadata = metadata or {}
        self._read_csv_kwargs = kwargs
        if structure is None:
            table = dask.dataframe.read_csv(
                self._file_paths[0], **self._read_csv_kwargs
            )
            structure = TableStructure.from_dask_dataframe(table)
            structure.npartitions = len(self._file_paths)
        self._structure = structure
        self.specs = list(specs or [])

    @classmethod
    def from_catalog(
        cls,
        data_source: DataSource[TableStructure],
        node: Node,
        /,
        **kwargs: Optional[Any],
    ) -> "CSVAdapter":
        return init_adapter_from_catalog(cls, data_source, node, **kwargs)  # type: ignore

    @classmethod
    def from_uris(
        cls,
        *data_uris: str,
        **kwargs: Optional[Any],
    ) -> "CSVAdapter":
        return cls(data_uris, **kwargs)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._structure.columns!r})"

    def metadata(self) -> JSON:
        return self._metadata

    @classmethod
    def init_storage(
        cls,
        storage: Storage,
        data_source: DataSource[TableStructure],
        path_parts: List[str],
    ) -> DataSource[TableStructure]:
        """Initialize partitioned CSV storage

        Parameters
        ----------
        data_uri : str
            location of the dataset, should point to a folder in which partitioned csv files will be created
        structure : TableStructure
            description of the data structure

        Returns
        -------
            list of assets with each element corresponding to individual partition files
        """
        data_source = copy.deepcopy(data_source)  # Do not mutate caller input.
        data_uri = storage.uri + "".join(
            f"/{quote_plus(segment)}" for segment in path_parts
        )
        directory = path_from_uri(data_uri)
        directory.mkdir(parents=True, exist_ok=True)
        assets = [
            Asset(
                data_uri=f"{data_uri}/partition-{i}.csv",
                is_directory=False,
                parameter="data_uris",
                num=i,
            )
            for i in range(data_source.structure.npartitions)
        ]
        data_source.assets.extend(assets)
        return data_source

    def append_partition(
        self, data: Union[dask.dataframe.DataFrame, pandas.DataFrame], partition: int
    ) -> None:
        """Append data to an existing partition

        Parameters
        ----------
        data : dask.dataframe.DataFrame or pandas.DataFrame
            data to be appended
        partition : int
            index of the partition to be appended to

        """
        uri = self._file_paths[partition]
        data.to_csv(uri, index=False, mode="a", header=False)

    def write_partition(
        self, data: Union[dask.dataframe.DataFrame, pandas.DataFrame], partition: int
    ) -> None:
        """Write data to a new partition or overwrite an existing one

        Parameters
        ----------
        data : dask.dataframe.DataFrame or pandas.DataFrame
            data to be appended
        partition : int
            index of the partition to be appended to

        """
        uri = self._file_paths[partition]
        data.to_csv(uri, index=False)

    def write(self, data: Union[dask.dataframe.DataFrame, pandas.DataFrame]) -> None:
        """Default writing function to a dataset with a single partition

        Parameters
        ----------
        data : dask.dataframe.DataFrame or pandas.DataFrame
            data to be written

        """
        if self.structure().npartitions != 1:
            raise NotImplementedError
        uri = self._file_paths[0]
        data.to_csv(uri, index=False)

    def read(self, fields: Optional[List[str]] = None) -> dask.dataframe.DataFrame:
        """

        Parameters
        ----------
        fields :

        Returns
        -------

        """
        dfs = [
            self.read_partition(i, fields=fields) for i in range(len(self._file_paths))
        ]

        return dask.dataframe.concat(dfs, axis=0)

    def read_partition(
        self,
        indx: int,
        fields: Optional[List[str]] = None,
    ) -> dask.dataframe.DataFrame:
        """Read a single partition

        Parameters
        ----------
        indx : int
            index of the partition to read
        fields :

        Returns
        -------

        """

        df = dask.dataframe.read_csv(self._file_paths[indx], **self._read_csv_kwargs)

        if fields is not None:
            df = df[fields]

        return df.compute()

    def structure(self) -> TableStructure:
        return self._structure

    def get(self, key: str) -> Union[ArrayAdapter, None]:
        """

        Parameters
        ----------
        key :

        Returns
        -------

        """
        if key not in self.structure().columns:
            return None
        return ArrayAdapter.from_array(self.read([key])[key].values)

    def generate_data_sources(
        self,
        mimetype: str,
        dict_or_none: Callable[[TableStructure], Dict[str, str]],
        item: Union[str, Path],
        is_directory: bool,
    ) -> List[DataSource[TableStructure]]:
        """

        Parameters
        ----------
        mimetype :
        dict_or_none :
        item :
        is_directory :

        Returns
        -------

        """
        return [
            DataSource(
                structure_family=StructureFamily.table,
                mimetype=mimetype,
                structure=self.structure(),
                parameters={},
                management=Management.external,
                assets=[
                    Asset(
                        data_uri=ensure_uri(item),
                        is_directory=is_directory,
                        parameter="data_uris",  # <-- PLURAL!
                        num=0,  # <-- denoting that the Adapter expects a list, and this is the first element
                    )
                ],
            )
        ]

    def __getitem__(self, key: str) -> ArrayAdapter:
        """Get an ArrayAdapter for a single column

        Parameters
        ----------
        key : str
            column name to get

        Returns
        -------
        An array adapter corresponding to a single column in the table.
        """
        # Must compute to determine shape.
        return ArrayAdapter.from_array(self.read([key])[key].values)

    def items(self) -> Iterator[Tuple[str, ArrayAdapter]]:
        """Iterator over table columns

        Returns
        -------
        Tuples of column names and corresponding ArrayAdapters
        """
        yield from (
            (key, ArrayAdapter.from_array(self.read([key])[key].values))
            for key in self._structure.columns
        )


class CSVArrayAdapter(ArrayAdapter):
    """Adapter for array-type data stored as partitioned csv files"""

    @classmethod
    def from_catalog(
        cls,
        data_source: DataSource[ArrayStructure],
        node: Node,
        /,
        **kwargs: Optional[Any],
    ) -> "CSVArrayAdapter":
        """Adapter for partitioned array data stored as a sequence of csv files

        Parameters
        ----------
        data_source :
        node :
        kwargs : dict
            any keyword arguments that can be passed to the pandas.read_csv function, e.g. names, sep, dtype, etc.
        """

        # Load the array lazily with Dask
        file_paths = [path_from_uri(ast.data_uri) for ast in data_source.assets]
        structure = data_source.structure
        dtype_numpy = structure.data_type.to_numpy_dtype()
        nrows = kwargs.pop("nrows", None)  # dask doesn't accept nrows
        _kwargs = {"dtype": dtype_numpy, "header": None}
        _kwargs.update(kwargs)
        ddf = dask.dataframe.read_csv(file_paths, **_kwargs)
        chunks_0: tuple[int, ...] = structure.chunks[
            0
        ]  # chunking along the rows dimension (when not stackable)
        if not dtype_numpy.isbuiltin:
            # Structural np dtype (0) -- return a records array
            # NOTE: dask.DataFrame.to_records() allows one to pass `index=False` to drop the index column, but as
            #       of desk ver. 2024.2.1 it seems broken and doesn't do anything. Instead, we set an index to any
            #       (first) column in the df to prevent it from creating an extra one.
            array = ddf.set_index(ddf.columns[0]).to_records(lengths=chunks_0)
        else:
            # Simple np dtype (1 or 2) -- all fields have the same type -- return a usual array
            array = ddf.to_dask_array(lengths=chunks_0)

        # Possibly extend or cut the table according the nrows parameter
        if nrows is not None:
            # TODO: this pulls all the data and can take long to compute. Instead, we can open the files and
            #       iterate over the rows directly, which is about 4-5 times faster for 50K rows.
            #       Can also just .compute() and return a np array instead
            nrows_actual = len(ddf)
            if nrows > nrows_actual:
                padding = dask.array.zeros_like(
                    array, shape=(nrows - nrows_actual, *array.shape[1:])
                )
                array = dask.array.append(array[:nrows_actual, ...], padding, axis=0)
            else:
                array = array[:nrows, ...]

            array = array.reshape(structure.shape).rechunk(structure.chunks)

        return cls(
            array,
            structure,
            metadata=node.metadata_,
            specs=node.specs,
        )

    @classmethod
    def from_uris(
        cls,
        *data_uris: str,
        **kwargs: Optional[Any],
    ) -> "CSVArrayAdapter":
        file_paths = [path_from_uri(uri) for uri in data_uris]
        array = dask.dataframe.read_csv(
            file_paths, header=None, **kwargs
        ).to_dask_array()
        structure = ArrayStructure.from_array(array)

        return cls(array, structure)



import asyncio
from pathlib import Path
from typing import Optional

import typer

from ._serve import serve_catalog

catalog_app = typer.Typer(no_args_is_help=True)
# Support both `tiled serve catalog` and `tiled catalog serve` as synonyms
# because I cannot decide which is right.
catalog_app.command("serve")(serve_catalog)
DEFAULT_SQLITE_CATALOG_FILENAME = "catalog.db"


@catalog_app.command("init")
def init(
    database: str = typer.Argument(
        Path.cwd() / DEFAULT_SQLITE_CATALOG_FILENAME, help="A filepath or database URI"
    ),
    if_not_exists: bool = typer.Option(
        False,
        help=(
            "By default, it is an error if a database is already initialized."
            "Set this flag to be permissive and return without an error."
        ),
    ),
):
    """
    Initialize a database as a Tiled Catalog.

    Examples:

    # Using a simple local file as an embedded "database" (SQLite)
    tiled init catalog.db
    tiled init path/to/catalog.db
    tiled init sqlite:////path/to/catalog.db

    # Using a client/serve database engine (PostgreSQL)
    tiled init postgresql://username:password@localhost/database_name:5432
    """
    from sqlalchemy.ext.asyncio import create_async_engine

    from ..alembic_utils import UninitializedDatabase, check_database, stamp_head
    from ..catalog.alembic_constants import ALEMBIC_DIR, ALEMBIC_INI_TEMPLATE_PATH
    from ..catalog.core import ALL_REVISIONS, REQUIRED_REVISION, initialize_database
    from ..utils import ensure_specified_sql_driver

    database = ensure_specified_sql_driver(database)

    async def do_setup():
        engine = create_async_engine(database)
        redacted_url = engine.url._replace(password="[redacted]")
        try:
            await check_database(engine, REQUIRED_REVISION, ALL_REVISIONS)
        except UninitializedDatabase:
            # Create tables and stamp (alembic) revision.
            typer.echo(
                f"Database {redacted_url} is new. Creating tables.",
                err=True,
            )
            await initialize_database(engine)
            typer.echo("Database initialized.", err=True)
        else:
            if not if_not_exists:
                typer.echo(
                    f"Database at {redacted_url} is already initialized.", err=True
                )
                raise typer.Abort()
        finally:
            await engine.dispose()

    asyncio.run(do_setup())
    stamp_head(ALEMBIC_INI_TEMPLATE_PATH, ALEMBIC_DIR, database)


@catalog_app.command("upgrade-database")
def upgrade_database(
    database_uri: str,
    revision: Optional[str] = typer.Argument(
        None,
        help="The ID of a revision to upgrade to. By default, upgrade to the latest one.",
    ),
):
    """
    Upgrade the catalog database schema to the latest version.
    """
    import asyncio

    from sqlalchemy.ext.asyncio import create_async_engine

    from ..alembic_utils import get_current_revision, upgrade
    from ..catalog.alembic_constants import ALEMBIC_DIR, ALEMBIC_INI_TEMPLATE_PATH
    from ..catalog.core import ALL_REVISIONS
    from ..utils import ensure_specified_sql_driver

    database_uri = ensure_specified_sql_driver(database_uri)

    async def do_setup():
        engine = create_async_engine(database_uri)
        redacted_url = engine.url._replace(password="[redacted]")
        current_revision = await get_current_revision(engine, ALL_REVISIONS)
        await engine.dispose()
        if current_revision is None:
            # Create tables and stamp (alembic) revision.
            typer.echo(
                f"Database {redacted_url} has not been initialized. Use `tiled catalog init`.",
                err=True,
            )
            raise typer.Abort()

    asyncio.run(do_setup())
    upgrade(ALEMBIC_INI_TEMPLATE_PATH, ALEMBIC_DIR, database_uri, revision or "head")


@catalog_app.command("downgrade-database")
def downgrade_database(
    database_uri: str,
    revision: str = typer.Argument(..., help="The ID of a revision to downgrade to."),
):
    """
    Upgrade the catalog database schema to the latest version.
    """
    import asyncio

    from sqlalchemy.ext.asyncio import create_async_engine

    from ..alembic_utils import downgrade, get_current_revision
    from ..catalog.alembic_constants import ALEMBIC_DIR, ALEMBIC_INI_TEMPLATE_PATH
    from ..catalog.core import ALL_REVISIONS
    from ..utils import ensure_specified_sql_driver

    database_uri = ensure_specified_sql_driver(database_uri)

    async def do_setup():
        engine = create_async_engine(database_uri)
        redacted_url = engine.url._replace(password="[redacted]")
        current_revision = await get_current_revision(engine, ALL_REVISIONS)
        if current_revision is None:
            # Create tables and stamp (alembic) revision.
            typer.echo(
                f"Database {redacted_url} has not been initialized. Use `tiled catalog init`.",
                err=True,
            )
            raise typer.Abort()

    asyncio.run(do_setup())
    downgrade(ALEMBIC_INI_TEMPLATE_PATH, ALEMBIC_DIR, database_uri, revision)
